(** Inputs for the Hoare-Triple verifier extracted in Caml 

NB: Coq notations gives use a cheap parser for PL language...

*)

Require Import ZArith.
Require Import DemoPLVerifier.
Import NumC.
Import AffCond.ZCond.
Import Term.
Import String.

Open Scope string_scope.
Open Scope Z_scope.
Open Scope cond_scope.
Open Scope term_scope.
Open Scope statement_scope.

Local Notation var:=ProgVar.PVar.t.

Module Type VariableBasis.

  Parameter x q r a: var.

End VariableBasis.

Module Examples(B:VariableBasis).

Import B.

Coercion Var: var >-> term.

Definition idZ (x:Z) : ZNum.t := x.
Coercion idZ: Z >-> ZNum.t.
Coercion Cte: ZNum.t >-> term.

Definition idTerm (x:term): AffTerm.ZTerm.term := x.
Coercion idTerm: term >-> AffTerm.ZTerm.term.

(* a ::= abs(x) -; q ::= a / 2 -; r ::= a % 2 *)
Definition div2 (inv: cond):=
  If (0 <= x) Then
    a ::= x
  Else
    a ::= Opp x
  Fi -;
  r ::= a -;
  q ::= 0 -;
  While (2 <= r) Do
    q ::= q+1 -;
    r ::= r-2
  Inv (a=q*2 + r /\ 0 <= r /\ 0 <= q /\ inv)
  Done.
  

Open Scope list_scope.

(* Hoare-Triple expected to be accepted *)
Definition tests_ok : list (string * hoareTriple) :=
    ("div2_ok0", {| pre:=true ; 
                    prog:=div2 true; 
                    post:=(0 <= q /\ 0 <= r /\ r <= 1) |})
    ::("div2_ok1", {| pre:=(-100 < x /\ x < 100) ; 
                      prog:=div2 (a < 100);
                      post:=(0 <= q /\ q < 50 /\ 0 <= r /\ r <= 1) |} )
    ::("div2_ok2", {| pre:=(x=0) ;
                      prog:=div2 (a = 0);
                      post:=(q=0 /\ r=0) |})
    ::("div2_ok3", {| pre:=(x=99) ; 
                      prog:=div2 (a = 99);
                      post:=(49 <= q /\ q < 50 /\ q <> 50 /\ 0 <= r /\ r <= 1) |} )
    ::nil.

(* Hoare-Triple expected to be rejected (because either invalid or incompleteness) *)
Definition tests_ko : list (string * hoareTriple) :=
    ("div2_ko_init", {| pre:=true ; 
                        prog:=div2 (a < 100); 
                        post:=(q <= 50) |} )
    ::("div2_ko_post", {| pre:=(-100 < x /\ x < 100) ; 
                          prog:=div2 true; 
                          post:=(0 <= q /\ q <= 50 /\ 0 <= r /\ r <= 1) |} )
    ::("div2_ko_preserv", {| pre:=(-100 < x /\ x < 100) ; 
                             prog:=div2 (q <= 50); 
                             post:=(q <= 50) |} )
    ::("div2_ko_integer_incompletude", 
                     {| pre:=(x=99) ; 
                        prog:=div2 (a=99);
                        post:=(q=49 /\ r=1) |} )
   ::nil.

  
End Examples.

