Require Export Impure.

(** model of pure computations (used for extraction !) *)
Module PureImpureMonad: ImpureMonad.
   
   Definition imp (A:Type) := A.
   Definition pure {A:Type} (a:A) := a.
   Definition bind {A B:Type} (k1: A) (k2: A -> B) := k2 k1.

   Definition mayReturn {A:Type} (a b:A): Prop := a=b.

   Lemma mayReturn_pure (A:Type) (a b:A): (mayReturn (pure a) b) -> a=b.
   Proof.
     intuition.
   Qed.

   Lemma mayReturn_bind (A B:Type) (k1:imp A) (k2: A -> imp B) (b:B):
         (mayReturn (bind k1 k2) b) -> exists (a:A), (mayReturn k1 a) /\ (mayReturn (k2 a) b).
   Proof.
     firstorder.
   Qed.


  Definition impeq {A} := @eq A.

  Instance impeq_equiv A: Equivalence (@impeq A).
  Proof.
    unfold impeq; constructor 1; intro x; congruence.
  Qed.

  Instance bind_eq_compat A B: Proper ((@impeq A) ==> (pointwise_relation A (@impeq B)) ==> (@impeq B)) (@bind A B).
  Proof.
    unfold impeq, bind; intros x y; congruence.
  Qed.

  Instance mayReturn_eq_compat A: Proper ((@impeq A) ==> Logic.eq ==> iff) (@mayReturn A).
  Proof.
    unfold impeq, mayReturn, iff; intros k1 k2 H x y H0; firstorder congruence.
  Qed.

  Lemma impeq_bind_pure_l (A B:Type) (k: A -> imp B) (a:A):
     (impeq (bind (pure a) k) (k a)).
  Proof.
    unfold impeq, pure, bind; simpl; auto.
  Qed.

  Lemma impeq_bind_pure_r (A:Type) (k: imp A):
     (impeq (bind k (fun b => pure b)) k).
  Proof.
    unfold impeq, pure, bind. simpl; auto.
  Qed.

  Lemma impeq_bind_assoc (A B C:Type) (k1: imp A) (k2: A -> imp B) (k3: B -> imp C):
     (impeq (bind (bind k1 k2) (fun b => k3 b)) (bind k1 (fun a => bind (k2 a) (fun b => k3 b)))).
  Proof.
    unfold impeq, bind. simpl; auto.
  Qed.

End PureImpureMonad.

Module Config := ImpureMonadTheory(PureImpureMonad).
Export Config.

Global Opaque pure bind imp mayReturn impeq.
