type 'a mayFailT
=
	| Ok of 'a
	| Failed of string

let process: Dat.t mayFailT -> string -> Dat.t mayFailT
= fun dat m ->
	match dat with
	| Failed _ as err -> err
	| Ok dat ->
		match Dep.gather m with
		| Dep.Failed err -> Failed err
		| Dep.Ok deps ->
			match Dat.depends m deps dat with
			| Dat.Failed err -> Failed err
			| Dat.Ok dat1 -> Ok dat1

let getModList: string array -> string list mayFailT
= fun args ->
	let rec build l i =
		if i >= 1 then
			build ((Array.get args i)::l) (i - 1)
		else
			l
	in
	let argc = Array.length args in
	if argc < 2 then
		Failed "need at least a file name"
	else
		let files = build [] (argc -1) in
		Ok files

let main: string array -> unit
= fun args ->
	match getModList args with
	| Failed err -> failwith err
	| Ok files ->
		match List.fold_left process (Ok Dat.nil) files with
		| Failed err -> Pervasives.prerr_endline err
		| Ok dat ->
			match Dat.order dat with
			| Dat.Failed err -> (
				Pervasives.prerr_endline err;
				Pervasives.exit 1
			)
			| Dat.Ok l -> Pervasives.print_string (String.concat " " l)

let _ = main Sys.argv
