(* IneqSet *)
let x = Var.fromInt 1
let y = Var.fromInt 2
let z = Var.fromInt 3
let t1 = Var.fromInt 4
let t3 = Var.fromInt 5
let t2 = Var.fromInt 6

let nxt = Var.fromInt 7

let varPr: Var.t -> string
= fun _x ->
	let vars: (Var.t * string) list
	= [x, "x"; y, "y"; z, "z"; t1, "t1"; t2, "t2"; t3, "t3"]
	in
	try
		List.assoc _x vars
	with
	| Not_found -> "v" ^ (Var.pr _x)

let mkc t v c =
	Cstr.mk t (List.map (fun (n, x) -> (Nb.mk1 n, x)) v) (Nb.mk1 c)

let eq = mkc Cstr.Eq
let le = mkc Cstr.Le
let lt = mkc Cstr.Lt

let mask l =
	List.fold_left (fun m x -> Rtree.set None m x (Some x)) Rtree.Nil l

let optxpr =
	function
	| None -> "None"
	| Some x -> "Some " ^ (varPr x)

let mkN = List.map (fun (i, n) -> (i, Nb.mk1 n))

let propPr = function
	| IneqSet.Empty f -> Printf.sprintf "Empty: %s" (Cert.prF f)
	| IneqSet.Trivial -> "Trivial"
	| IneqSet.Implied (i, ce) -> Printf.sprintf "Implied %i: %s" i (Cert.prC ce)
	| IneqSet.Check c -> Printf.sprintf "Check: %s" (Cons.prExt varPr c)

let mkEqSet: (int * Cstr.t) list -> EqSet.t
= fun l ->
	match EqSet.addM EqSet.nil l with
	| EqSet.Added (s, _) -> s
	| EqSet.Bot _ -> failwith "IneqSet_t.mkEqSet"

let iset: (int * Cstr.t) list -> IneqSet.t
= fun l ->
	match IneqSet.addM nxt IneqSet.nil l with
	| IneqSet.Added (s, _) -> s
	| IneqSet.ImplicitEqs (_, _)
	| IneqSet.Contrad _ -> Pervasives.invalid_arg "IneqSet_t.iset"

let relEq: IneqSet.rel_t -> IneqSet.rel_t -> bool
= fun r1 r2 ->
	match r1, r2 with
	| IneqSet.NoIncl, IneqSet.NoIncl -> true
	| IneqSet.Incl l1, IneqSet.Incl l2 -> Cert.partialEq l1 l2
	| IneqSet.NoIncl, IneqSet.Incl _
	| IneqSet.Incl _, IneqSet.NoIncl -> false

let relPr: IneqSet.rel_t -> string
= function
	| IneqSet.NoIncl -> "NoIncl"
	| IneqSet.Incl l -> Printf.sprintf "Incl l with l:\n%s" (Cert.partialPr l)

(* IneqSet.synIncl *)
let synInclCheckTs: T.testT
=
	let chk (t, es, s, cIn, cOut) = fun state ->
		let eq c1 c2 =
			match c1, c2 with
			| IneqSet.Check c1, IneqSet.Check c2 ->
				((Cons.get_id c1) = (Cons.get_id c2)) &&
				Cons.isEq c1 c2 &&
				Cert.isEqF (Cons.get_f c1) (Cons.get_f c2)

			| _, _ -> false
		in
		let result = IneqSet.synIncl es s cIn in
		if eq (IneqSet.Check cOut) result then
			T.succeed state
		else
			T.fail t (propPr result) state
	in
	let c i c f = { Cons.id = i; Cons.c = c; Cons.f = Cert.mkF (mkN f) } in
	let tcs = [
		"subst0",
			mkEqSet [ 0, eq [1, y] 0 ], iset [],
			c 2 (le [1, x; 1, y] 0) [2, 1],
			c 2 (le [1, x] 0) [0, -1; 2, 1]
	] in
	T.suite "check" (List.map chk tcs)

let synInclTs: T.testT
= T.suite "synIncl" [synInclCheckTs]

(* IneqSet.addM *)
let addMTs: T.testT
=	let chk (nm, s, l, r)
	= fun st ->
		let ar = IneqSet.addM nxt s l in
		if IneqSet.meetEq r ar
		then T.succeed st
		else
			let estr = Printf.sprintf "expected:\n%s\ngot:\n%s\n"
				(IneqSet.meetPr varPr r) (IneqSet.meetPr varPr ar)
			in
			T.fail nm estr st
	in
	let tcs: (string * IneqSet.t * (int * Cstr.t) list * IneqSet.meetT) list
	= [
	"triv0", IneqSet.nil, [],
		IneqSet.Added (IneqSet.nil, ([], []));

	"triv1", iset [0, le [1, x] 0], [],
		IneqSet.Added (iset [0, le [1, x] 0],
		([0, Cert.Direct (Cert.mkF [0, Nb.u])],
		[0, Cert.Direct (Cert.mkF [0, Nb.u])]));

	"triv2", iset [], [0, le [] 0],
		IneqSet.Added (IneqSet.nil, ([], [0, Cert.Direct []]));

	"triv3", iset [0, le [1, x] 0], [1, le [] 0],
		IneqSet.Added (iset [0, le [1, x] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u])],
			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct []]));

	"concat0", iset [], [0, le [1, x] 0],
		IneqSet.Added (iset [0, le [1, x] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u])],
			[0, Cert.Direct (Cert.mkF [0, Nb.u])]));

	"concat1", iset [0, le [1, x] 0], [1, le [1, y] 0],
		IneqSet.Added (iset [
			0, le [1, x] 0; 
			1, le [1, y] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u])],

			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u])]));

	"eqs0", iset [0, le [1, x] 0], [1, le [-1, x] 0],
		IneqSet.ImplicitEqs (([
			Cons.mk 0 (le [1, x] 0);
			Cons.mk 1 (le [-1, x] 0)],
			[eq [1, x] 0, Cert.SplitEq (Cert.mkF [0, Nb.u], Cert.mkF [1, Nb.u]);
			eq [-1, x] 0, Cert.SplitEq (Cert.mkF [1, Nb.u], Cert.mkF [0, Nb.u])]),
			([0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u])],

			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u])]));

	"eqs1", iset [
			0, le [1, x; 1, y] 0;
			1, le [1, z] 0],
		[2, le [-1, x; -1, y] 0],
		IneqSet.ImplicitEqs (([
			Cons.mk 0 (le [1, x; 1, y] 0);
			Cons.mk 1 (le [1, z] 0);
			Cons.mk 2 (le [-1, x; -1, y] 0)],
			[eq [1, x; 1, y] 0,
				Cert.SplitEq (Cert.mkF [0, Nb.u], Cert.mkF [2, Nb.u]);
			eq [-1, x; -1, y] 0,
				Cert.SplitEq (Cert.mkF [2, Nb.u], Cert.mkF [0, Nb.u])]),
			([0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u]);
			2, Cert.Direct (Cert.mkF [2, Nb.u])],

			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u]);
			2, Cert.Direct (Cert.mkF [2, Nb.u])]));

	"eqs2", iset [0, le [1, x] 0], [1, le [-2, x] 0],
		IneqSet.ImplicitEqs (([
			Cons.mk 0 (le [1, x] 0);
			Cons.mk 1 (le [-2, x] 0)],
			[eq [1, x] 0,
				Cert.SplitEq (Cert.mkF [0, Nb.u], Cert.mkF [1, Nb.mk 2 1]);
			eq [-2, x] 0,
				Cert.SplitEq (Cert.mkF [1, Nb.u], Cert.mkF [0, Nb.mk1 2])]),

			([0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u])],

			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u])]));

	"eqs3", iset [
		0, le [1, x; -1, y] 0;
		1, le [1, y; -1, z] 0],
		[2, le [1, z; -1, x] 0],
		IneqSet.ImplicitEqs (([
			Cons.mk 0 (le [1, x; -1, y] 0);
			Cons.mk 1 (le [1, y; -1, z] 0);
			Cons.mk 2 (le [1, z; -1, x] 0)],
			[eq [1, x; -1, y] 0,
				Cert.SplitEq (Cert.mkF [0, Nb.u], Cert.mkF [1, Nb.u; 2, Nb.u]);
			eq [1, y; -1, z] 0,
				Cert.SplitEq (Cert.mkF [1, Nb.u], Cert.mkF [0, Nb.u; 2, Nb.u]);
			eq [-1, x; 1, z] 0,
				Cert.SplitEq (Cert.mkF [2, Nb.u], Cert.mkF [0, Nb.u; 1, Nb.u])]),
			([0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u]);
			2, Cert.Direct (Cert.mkF [2, Nb.u])],

			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u]);
			2, Cert.Direct (Cert.mkF [2, Nb.u])]));

	"eqs4", iset [
		0, le [1, x; -1, y] 0;
		1, le [1, y; -1, z] 0;
		2, le [1, t1; 1, x; -1, z; -1, t2] 0;
		3, le [1, t2; -1, t1] 0],
		[4, le [-1, x; 1, z] 0],
		IneqSet.ImplicitEqs (([
			Cons.mk 0 (le [1, x; -1, y] 0);
			Cons.mk 1 (le [1, y; -1, z] 0);
			Cons.mk 2 (le [1, t1; 1, x; -1, z; -1, t2] 0);
			Cons.mk 3 (le [1, t2; -1, t1] 0);
			Cons.mk 4 (le [-1, x; 1, z] 0)],
		[eq [1, x; -1, y] 0,
			Cert.SplitEq (Cert.mkF [0, Nb.u], Cert.mkF [1, Nb.u; 4, Nb.u]);
		eq [1, y; -1, z] 0,
			Cert.SplitEq (Cert.mkF [1, Nb.u], Cert.mkF [0, Nb.u; 4, Nb.u]);
		eq [-1, x; 1, z] 0,
			Cert.SplitEq (Cert.mkF [4, Nb.u], Cert.mkF [0, Nb.u; 1, Nb.u]);
		eq [1, t1; 1, x; -1, z; -1, t2] 0,
			Cert.SplitEq (Cert.mkF [2, Nb.u], Cert.mkF [3, Nb.u; 4, Nb.u]);
		eq [1, t2; -1, t1] 0,
			Cert.SplitEq (Cert.mkF [3, Nb.u], Cert.mkF [2, Nb.u; 4, Nb.u])]),

		([0, Cert.Direct (Cert.mkF [0, Nb.u]);
		1, Cert.Direct (Cert.mkF [1, Nb.u]);
		2, Cert.Direct (Cert.mkF [2, Nb.u]);
		3, Cert.Direct (Cert.mkF [3, Nb.u]);
		4, Cert.Direct (Cert.mkF [4, Nb.u])],

		[0, Cert.Direct (Cert.mkF [0, Nb.u]);
		1, Cert.Direct (Cert.mkF [1, Nb.u]);
		2, Cert.Direct (Cert.mkF [2, Nb.u]);
		3, Cert.Direct (Cert.mkF [3, Nb.u]);
		4, Cert.Direct (Cert.mkF [4, Nb.u])]));

	"synred0", iset [0, le [1, x] 0], [1, le [1, x] 0],
		IneqSet.Added (iset [0, le [1, x] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u])],
			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [0, Nb.u])]));

	"synred1", iset [0, le [1, x] 0], [1, le [1, x] 1],
		IneqSet.Added (iset [0, le [1, x] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u])],
			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [0, Nb.u])]));

	"synred2", iset [0, le [1, x] 0], [1, le [2, x] 0],
		IneqSet.Added (iset [0, le [1, x] 0],
			([0, Cert.Direct (Cert.mkF [0, Nb.u])],
			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [0, Nb.mk1 2])]));

	"synred3", iset [0, le [1, x] 1], [1, le [1, x] 0],
		IneqSet.Added (iset [1, le [1, x] 0],
			([1, Cert.Direct (Cert.mkF [1, Nb.u])],
			[0, Cert.Direct (Cert.mkF [1, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u])]));

	"red0", iset [
		0, le [1, x] 1;
		1, le [1, y] 2],
		[2, le [1, x; 2, y] 5],
		IneqSet.Added (iset [
			0, le [1, x] 1;
			1, le [1, y] 2],
			([0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u])],
			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u]);
			2, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.mk1 2])]));

	"red1", iset [
		0, le [1, x] 1;
		1, le [1, y] 2;
		2, le [1, x; -1, y] 2],
		[3, le [1, x; 2, y] 5],
		IneqSet.Added (iset [
			0, le [1, x] 1;
			1, le [1, y] 2;
			2, le [1, x; -1, y] 2],
			([0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u]);
			2, Cert.Direct (Cert.mkF [2, Nb.u])],
			[0, Cert.Direct (Cert.mkF [0, Nb.u]);
			1, Cert.Direct (Cert.mkF [1, Nb.u]);
			2, Cert.Direct (Cert.mkF [2, Nb.u]);
			3, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.mk1 2])]));

	"red2", iset [
		0, le [1, x] 1;
		1, le [2, x; 1, y] 3],
		[2, le [1, y] 1],
		IneqSet.Added (iset [
			0, le [1, x] 1;
			2, le [1, y] 1],
		([0, Cert.Direct (Cert.mkF [0, Nb.u]);
		2, Cert.Direct (Cert.mkF [2, Nb.u])],

		[0, Cert.Direct (Cert.mkF [0, Nb.u]);
		1, Cert.Direct (Cert.mkF [0, Nb.mk1 2; 2, Nb.u]);
		2, Cert.Direct (Cert.mkF [2, Nb.u])]));

	"contrad0", iset [], [0, le [] (-1)],
		IneqSet.Contrad (Cert.mkF [0, Nb.u]);

	"contrad1", iset [0, le [1, x] 0], [1, le [-1, x] (-1)],
		IneqSet.Contrad (Cert.mkF [0, Nb.u; 1, Nb.u]);

	"contrad2", iset [0, le [1, x] 0], [1, le [-2, x] (-2)],
		IneqSet.Contrad (Cert.mkF [0, Nb.u; 1, Nb.mk 2 1]);

	"contrad3", iset [
		0, le [1, x] 0;
		1, le [1, y] 1],
		[2, le [-1, x; -2, y] (-3)],
		IneqSet.Contrad (Cert.mkF [0, Nb.mk 2 1; 1, Nb.u; 2, Nb.mk 2 1])
	] in
	T.suite "addM" (List.map chk tcs)

(* IneqSet.pick *)
let pickTs: T.testT
=
	let chk (t, msk, x, l) = fun state ->
		let s = iset l in
		let x1 = IneqSet.pick msk s in
		if x = x1 then
			T.succeed state
		else
			T.fail t (optxpr x1) state
	in
	let tcs = [
		"nil0", mask [], None, [];
		"nil1", mask [x], None, [];
		"nil2", mask [y], None, [];
		"no0", mask [], None, [0, le [1, x] 0];
		"no1", mask [y], None, [0, le [1, x] 0];
		"no2", mask [x], None, [0, le [1, z] 0];
		"one0", mask [x], Some x, [0, le [1, x] 0];
		"one1", mask [x], Some x, [
			0, le [1, x] 0;
			1, le [1, y] 0 ];

		"one2", mask [x; y], Some x, [
			0, le [1, x] 0;
			1, le [1, x; 1, y] 0 ];

		"m0", mask [x; y], Some y, [
			0, le [1, x; 2, z] 0;
			1, le [1, x; 1, z] 0;
			2, le [-1, x; 1, y] 0;
			3, le [2, y; 1, z] 0 ]
	] in
	T.suite "pick" (List.map chk tcs)

(* IneqSet.subst *)
let substTs: T.testT
=
	let chk_id (t, x, i, i1, e, s, _) = fun state ->
		let s1 = iset s in
		let (i2, _) = IneqSet.subst nxt EqSet.nil x i e s1 in
		if i1 = i2 then
			T.succeed state
		else
			T.fail t (string_of_int i2) state
	in
	let chk_res (t, x, i, _, e, s, s1) = fun state ->
		let s2 = iset s in
		let s3 = iset s1 in
		let (_, s4) = IneqSet.subst nxt EqSet.nil x i e s2 in
		if IneqSet.isEq s3 s4 then
			T.succeed state
		else
			T.fail t (IneqSet.prExt varPr s4) state
	in
	let c i c = Cons.mk i c in
	let tcs = [
		"nil0", x, 1, 1, c 0 (eq [1, x] 0), [], [];
		"other0", x, 1, 2, c 0 (eq [1, x] 0), [
			1, le [1, y] 0
		], [
			1, le [1, y] 0 ];

		"other1", y, 1, 3, c 0 (eq [1, y] 0), [
			1, le [1, x] 0;
			2, le [1, x; 1, z] 0
		], [
			1, le [1, x] 0;
			2, le [1, x; 1, z] 0 ];

		"one0", x, 1, 2, c 0 (eq [1, x] 0), [
			1, le [1, x; 1, y] 0;
		], [
			1, le [1, y] 0 ];

		"mixed0", x, 1, 3, c 0 (eq [1, x] 0), [
			1, le [1, x; 1, y] 0;
			2, le [1, z] 0
		], [
			1, le [1, y] 0;
			2, le [1, z] 0 ];

		"simpl0", x, 1, 2, c 0 (eq [1, x] 0), [
			1, le [1, x; 1, y] 0;
			2, le [1, y] 0
		], [
			1, le [1, y] 0 ]
	] in
	T.suite "subst" [
		T.suite "id" (List.map chk_id tcs);
		T.suite "res" (List.map chk_res tcs)
	]

(* IneqSet.fmElim *)
let fmElimTs: T.testT
=
	let chk_id (t, x, i, i1, s, _) = fun state ->
		let s1 = iset s in
		let (i2, _) = IneqSet.fmElim nxt EqSet.nil x i s1 in
		if i1 = i2 then
			T.succeed state
		else
			T.fail t (string_of_int i2) state
	in
	let chk_res (t, x, i, _, s, s1) = fun state ->
		let s2 = iset s in
		let s3 = iset s1 in
		let (_, s4) = IneqSet.fmElim nxt EqSet.nil x i s2 in
		if IneqSet.isEq s3 s4 then
			T.succeed state
		else
			T.fail t (IneqSet.prExt varPr s4) state
	in
	let tcs = [
		"nil0", x, 1, 1, [], [];
		"other0", x, 1, 2, [
			1, le [1, y] 0
		], [
			1, le [1, y] 0 ];

		"plus0", x, 1, 2, [
			1, le [1, x] 0;
			2, le [1, y] 0
		], [
			1, le [1, y] 0 ];

		"plus1", x, 1, 2, [
			1, le [1, x] 0;
			2, le [1, x; 1, y] 0;
			3, le [1, y] 0
		], [
			2, le [1, y] 0 ];

		"plus2", x, 1, 1, [
			1, le [1, x] 0
		], [];

		"neg0", x, 1, 2, [
			1, le [-1, x] 0;
			2, le [1, y] 0
		], [
			1, le [1, y] 0 ];

		"m0", x, 1, 2, [
			1, le [1, x; 1, y] 0;
			2, le [-1, x] 0
		], [
			1, le [1, y] 0
		];

		"m1", x, 1, 1, [
			1, le [1, x] 1;
			2, le [-1, x] 0
		], [];

		"m2", x, 1, 3, [
			1, le [1, x; 1, y] 0;
			2, le [1, x; 1, z] 0;
			3, le [-1, x] 0
		], [
			1, le [1, y] 0;
			2, le [1, z] 0 ];

		"mixed0", x, 1, 3, [
			1, le [-1, x] 0;
			2, le [1, x; 1, y] 0;
			3, le [1, z] 0
		], [
			1, le [1, y] 0;
			2, le [1, z] 0 ]
	] in
	T.suite "fmElim" [
		T.suite "id" (List.map chk_id tcs);
		T.suite "res" (List.map chk_res tcs)
	]

(* IneqSet.fmElimM *)
let fmElimMTs: T.testT
=
	let chk_id (t, i, i1, msk, cList, _) = fun state ->
		let s = iset cList in
		let (i2, _) = IneqSet.fmElimM nxt EqSet.nil i msk s in
		if i1 = i2 then
			T.succeed state
		else
			T.fail t (string_of_int i2) state
	in
	let chk_res (t, i, _, msk, cList, cList1) =fun state ->
		let s = iset cList in
		let s1 = iset cList1 in
		let (_, s2) = IneqSet.fmElimM nxt EqSet.nil i msk s in
		if IneqSet.isEq s1 s2 then
			T.succeed state
		else
			T.fail t (IneqSet.prExt varPr s2) state
	in
	let tcs = [
		"nil0", 0, 0, mask [], [], [];
		"nil1", 0, 0, mask [x], [], [];
		"nil2", 0, 1, mask [], [
			1, le [1, x] 0
		], [
			1, le [1, x] 0 ];

		"two0", 0, 0, mask [x; y], [
			1, le [1, x] 0
		], [];

		"two1", 0, 0, mask [x; y], [
			1, le [1, x] 0;
			2, le [-1, x; 1, y] 0
		], [];

		"two2", 0, 1, mask [x; y], [
			1, le [1, x] 0;
			2, le [-1, x; 1, y] 0;
			3, le [-1, y; 1, z] 0
		], [
			1, le [1, z] 0 ]
	] in
	T.suite "fmElimM" [
		T.suite "id" (List.map chk_id tcs);
		T.suite "res" (List.map chk_res tcs)
	]

(* IneqSet.isIncl *)
let isInclTs: T.testT
=
	let chk (name, es, s1, s2, expected) = fun state ->
		let actual = IneqSet.isIncl nxt es s1 s2 in
		if relEq actual expected then
			T.succeed state
		else
			let err = Printf.sprintf "expected:\n%s\nactual:\n%s" (relPr expected) (relPr actual) in
			T.fail name err state
	in
	let tcs = [
		"simpl0", EqSet.nil, iset [
			0, le [1, x] 1
		], iset [
			0, le [1, x] 2 ],
		IneqSet.Incl [0, Cert.Direct (Cert.mkF [0, Nb.u])];

		"simpl1", EqSet.nil, iset [
			0, le [1, x] 2
		], iset [
			0, le [1, x] 1 ],
		IneqSet.NoIncl;

		"comb0", EqSet.nil, iset [
			1, le [1, y] 1;
			2, le [1, z] 1
		], iset [
			0, le [1, y; 1, z] 2 ],
		IneqSet.Incl [0, Cert.Direct (Cert.mkF [1, Nb.u; 2, Nb.u])];

		"substSimpl0", mkEqSet [0, eq [1, x] 0], iset [
			1, le [1, y] 1
		], iset [
			0, le [1, x; 1, y] 1 ],
		IneqSet.Incl [0, Cert.Direct (Cert.mkF [1, Nb.u; 0, Nb.u])];

		"substTriv0", mkEqSet [0, eq [1, x] 0], iset [], iset [
			0, le [-1, x] 0 ],
		IneqSet.Incl [0, Cert.Direct (Cert.mkF [0, Nb.negU])];

		"substTriv1", mkEqSet [0, eq [1, x] 0], iset [
			1, le [1, y] 0
		], iset [
			0, le [1, y] 0;
			1, le [-1, x] 0 ],
		IneqSet.Incl [0, Cert.Direct
			(Cert.mkF [1, Nb.u]); 1, Cert.Direct (Cert.mkF [0, Nb.negU])];

		"subst_comb0", mkEqSet [0, eq [1, z] 0], iset [
			1, le [-1, x; 1, y] 5;
			2, le [-1, x; -1, y] (-7)
		], iset [0, le [-1, x; 1, z] 0],
		IneqSet.Incl [0, Cert.Direct
			(Cert.mkF [0, Nb.u; 1, Nb.mk 2 1; 2, Nb.mk 2 1])]
	] in
	T.suite "isIncl" (List.map chk tcs)

(* IneqSet.fromUnsat *)
let fromUnsatKoTs: T.testT
=
	let chk (t, i, w) = fun state ->
		try
			let _ = IneqSet.fromUnsat (mkN w) i in
			T.fail t "exception not raised" state
		with
		| Invalid_argument "IneqSet.fromUnsat" -> T.succeed state
		| _ -> T.fail t "unexpected exception raised" state
	in
	let tcs = [
		"empty0", 1, [];
		"noI0", 1, [0, 1];
		"noI1", 1, [0, 1; 2, 3];
		"many0", 0, [0, 1; 0, 2];
		"many1", 0, [1, 2; 0, 1; 2, -5; 0, 2];
		"bad0", 1, [0, 1; 1, 0; 2, 3]
	] in
	T.suite "ko" (List.map chk tcs)

let fromUnsatOkTs: T.testT
=
	let chk (t, i, w, f) = fun state ->
		let result = IneqSet.fromUnsat (mkN w) i in
		if Cert.isEqF (mkN f) result then
			T.succeed state
		else
			T.fail t (Cert.prF result) state
	in
	let tcs = [
		"unit0", 1, [1, 1; 2, 1], [2, 1];

		(* normalization *)
		"unit1", 1, [1, 2; 2, 4], [2, 2];
		"unit2", 1, [1, -1; 3, 2], [3, 2]; (* XXX: Is this really what we want? *)

		(* order *)
		"unit3", 1, [2, 3; 1, 1], [2, 3];
		"unit4", 2, [4, 2; 2, 2], [4, 1];

		"m0", 2, [1, 2; 4, 2; 2, 1], [1, 2; 4, 2];
		"m1", 2, [1, 2; 4, 2; 2, 2], [1, 1; 4, 1]

	] in
	T.suite "ok" (List.map chk tcs)

let fromUnsatTs: T.testT
= T.suite "fromUnsat" [fromUnsatKoTs; fromUnsatOkTs]

let ts: T.testT
= T.suite "IneqSet" [synInclTs; addMTs; pickTs; substTs; fmElimTs; fmElimMTs; isInclTs; fromUnsatTs]
