type sign_t = P | Z | N

let sign a = if a < 0 then N else if a > 0 then P else Z
let pr = function P -> "P" | Z -> "Z" | N -> "N"

(* Nb.cmp *)
let cmpTs: T.testT
=
	let chk (t, n1, n2, s) = fun state ->
		let r = sign (Nb.cmp n1 n2) in
		if r = s then
			T.succeed state
		else
			let e = "Nb.cmp " ^ (Nb.pr n1) ^ " " ^ (Nb.pr n2) ^ ": " ^ (pr r) in
			T.fail t e state
	in
	let tcs = [
		"zz0", Nb.z, Nb.z, Z;
		"zu0", Nb.z, Nb.u, N;
		"zu1", Nb.z, Nb.negU, P;
		"uz0", Nb.u, Nb.z, P;
		"uz1", Nb.negU, Nb.z, N;
		"uu0", Nb.u, Nb.u, Z;
		"uu1", Nb.u, Nb.negU, P;
		"uu2", Nb.negU, Nb.u, N;
		"pz0", Nb.mk1 2, Nb.z, P;
		"zp0", Nb.z, Nb.mk1 2, N;
		"pp0", Nb.mk1 2, Nb.mk1 2, Z;
		"pp1", Nb.mk1 2, Nb.mk1 3, N;
		"pp2", Nb.mk1 3, Nb.mk1 2, P;
		"nn0", Nb.mk1 (-2), Nb.mk1 (-2), Z;
		"nn1", Nb.mk1 (-2), Nb.mk1 (-3), P;
		"nn2", Nb.mk1 (-3), Nb.mk1 (-2), N;
		"pn0", Nb.mk1 2, Nb.mk1 (-2), P;
		"np0", Nb.mk1 (-2), Nb.mk1 2, N
	] in
	T.suite "cmp" (List.map chk tcs)

(* Nb.cmpz *)
let cmpzTs: T.testT
=
	let chk (t, n, s) = fun state ->
		let r = sign (Nb.cmpz n) in
		if r = s then
			T.succeed state
		else
			let e = "Nb.cmpz " ^ (Nb.pr n) ^ ": " ^ (pr r) in
			T.fail t e state
	in
	let tcs = [
		"z", Nb.z, Z;
		"u0", Nb.u, N;
		"u1", Nb.negU, P;
		"p0", Nb.mk1 3, N;
		"p1", Nb.mk 10 2, N;
		"n0", Nb.mk1 (-4), P
	] in
	T.suite "cmpz" (List.map chk tcs)

(* Nb.neg *)
let negTs: T.testT
=
	let chk_res (t, n1, n2) = fun state ->
		let r = Nb.neg n1 in
		if Nb.cmp n2 r = 0 then
			T.succeed state
		else
			let e = "Nb.neg " ^ (Nb.pr n1) ^ " = " ^ (Nb.pr r) in
			T.fail t e state
	in
	let chk_comp (t, n1, _) = fun state ->
		let r = Nb.neg (Nb.neg n1) in
		if Nb.cmp n1 r = 0 then
			T.succeed state
		else
			let e = "Nb.neg (Nb.neg " ^ (Nb.pr n1) ^ ") = " ^ (Nb.pr r) in
			T.fail t e state
	in
	let tcs = [
		"z", Nb.z, Nb.z;
		"u", Nb.u, Nb.negU;
		"p0", Nb.mk1 2, Nb.mk1 (-2);
		"p1", Nb.mk1 5, Nb.mk1 (-5);
		"n0", Nb.negU, Nb.u;
		"n1", Nb.mk1 (-15), Nb.mk1 15
	] in
	T.suite "neg" [
		T.suite "res" (List.map chk_res tcs);
		T.suite "comp" (List.map chk_comp tcs)
	]

(* Nb.inv *)
(*
let invKoTs: T.testT
= fun state ->
	let r = Nb.inv Nb.z in
	if not (Nb.isReal r) then
		T.succeed state
	else
		let e = "Nb.inv Nb.z = " ^ (Nb.pr r) in
		T.fail "ko" e state
*)

let invOkTs: T.testT
=
	let chk_res (t, n1, n2) = fun state ->
		let n = Nb.inv n1 in
		if Nb.cmp n2 n = 0 then
			T.succeed state
		else
			let e = "Nb.inv " ^ (Nb.pr n1) ^ " = " ^ (Nb.pr n) in
			T.fail t e state
	in
	let chk_comp (t, n1, _) = fun state ->
		let n = Nb.inv (Nb.inv n1) in
		if Nb.cmp n1 n = 0 then
			T.succeed state
		else
			let e = "Nb.inv (Nb.inv " ^ (Nb.pr n1) ^ ") = " ^ (Nb.pr n) in
			T.fail t e state
	in
	let tcs = [
		"u0", Nb.u, Nb.u;
		"u1", Nb.negU, Nb.negU;
		"p0", Nb.mk1 2, Nb.mk 2 1;
		"n0", Nb.mk1 (-2), Nb.mk 2 (-1)
	] in
	T.suite "ok" [
		T.suite "res" (List.map chk_res tcs);
		T.suite "comp" (List.map chk_comp tcs)
	]

(* DM note: invKoTs does a divide by zero in F# *)
let invTs: T.testT
= T.suite "inv" [(* invKoTs; *) invOkTs]

(* Nb.add *)
let addTs: T.testT
=
	let chk_res (t, n1, n2, n) = fun state ->
		let r = Nb.add n1 n2 in
		if Nb.cmp n r = 0 then
			T.succeed state
		else
			let e = (Nb.pr n1) ^ " + " ^ (Nb.pr n2) ^ " = " ^ (Nb.pr r) in
			T.fail t e state
	in
	let chk_sym (t, n1, n2, _) = fun state ->
		let r1 = Nb.add n1 n2 in
		let r2 = Nb.add n2 n1 in
		if Nb.cmp r1 r2 = 0 then
			T.succeed state
		else
			let e1 = (Nb.pr n1) ^ " + " ^ (Nb.pr n2) ^ " = " ^ (Nb.pr r1) in
			let e2 = (Nb.pr n2) ^ " + " ^ (Nb.pr n1) ^ " = " ^ (Nb.pr r2) in
			T.fail t (e1 ^ ", " ^ e2) state
	in
	let tcs = [
		"zz", Nb.z, Nb.z, Nb.z;
		"u0", Nb.z, Nb.u, Nb.u;
		"u1", Nb.u, Nb.u, Nb.mk1 2;
		"p0", Nb.mk1 2, Nb.u, Nb.mk1 3;
		"p1", Nb.mk1 10, Nb.mk1 15, Nb.mk1 25;
		"n0", Nb.negU, Nb.z, Nb.negU;
		"n1", Nb.negU, Nb.u, Nb.z;
		"n2", Nb.mk1 (-5), Nb.u, Nb.mk1 (-4)
	] in
	T.suite "add" [
		T.suite "res" (List.map chk_res tcs);
		T.suite "sym" (List.map chk_sym tcs)
	]

(* Nb.mult *)
let multTs: T.testT
=
	let chk_res (t, n1, n2, n) = fun state ->
		let r = Nb.mult n1 n2 in
		if Nb.cmp n r = 0 then
			T.succeed state
		else	
			let e = (Nb.pr n1) ^ " + " ^ (Nb.pr n2) ^ " = " ^ (Nb.pr r) in
			T.fail t e state
	in
	let chk_sym (t, n1, n2, _) = fun state ->
		let r1 = Nb.mult n1 n2 in
		let r2 = Nb.mult n2 n1 in
		if Nb.cmp r1 r2 = 0 then
			T.succeed state
		else	
			let e1 = (Nb.pr n1) ^ " + " ^ (Nb.pr n2) ^ " = " ^ (Nb.pr r1) in
			let e2 = (Nb.pr n2) ^ " + " ^ (Nb.pr n1) ^ " = " ^ (Nb.pr r2) in
			T.fail t (e1 ^ ", " ^ e2) state
	in
	let tcs = [
		"z0", Nb.z, Nb.z, Nb.z;
		"z1", Nb.z, Nb.u, Nb.z;
		"z2", Nb.z, Nb.mk1 2, Nb.z;
		"z3", Nb.z, Nb.mk1 (-2), Nb.z;
		"u0", Nb.u, Nb.u, Nb.u;
		"u1", Nb.u, Nb.mk1 2, Nb.mk1 2;
		"u2", Nb.u, Nb.mk1 (-2), Nb.mk1 (-2);
		"p0", Nb.mk1 2, Nb.mk1 3, Nb.mk1 6;
		"p1", Nb.mk1 10, Nb.mk1 15, Nb.mk1 150;
		"n0", Nb.negU, Nb.u, Nb.negU;
		"n1", Nb.negU, Nb.negU, Nb.u;
		"n2", Nb.mk1 (-5), Nb.mk1 5, Nb.mk1 (-25);
		"n2", Nb.mk1 (-5), Nb.mk1 (-5), Nb.mk1 25
	] in
	T.suite "mult" [
		T.suite "res" (List.map chk_res tcs);
		T.suite "sym" (List.map chk_sym tcs)
	]

(* Nb.div *)
(*
let divKoTs: T.testT
=
	let chk (t, n, d) = fun state ->
		let r = Nb.div n d in
		if not (Nb.isReal r) then
			T.succeed state
		else			
			let e = (Nb.pr r) ^ "is real" in
			T.fail t e state
	in
	let tcs = [
		"zz", Nb.z, Nb.z;
		"u0", Nb.u, Nb.z;
		"u1", Nb.negU, Nb.z;
		"p0", Nb.mk1 2, Nb.z;
		"n0", Nb.mk1 (-5), Nb.z
	] in
	T.suite "ko" (List.map chk tcs)
*)

let divOkTs: T.testT
=
	let chk (t, n, d, r) = fun state ->
		let r1 = Nb.div n d in
		if Nb.cmp r r1 = 0 then
			T.succeed state
		else	
			let e = "Nb.div " ^ (Nb.pr n) ^ " " ^ (Nb.pr d) ^ " = " ^ (Nb.pr r1) in
			T.fail t e state
	in
	let tcs = [
		"u0", Nb.z, Nb.u, Nb.z;
		"u1", Nb.z, Nb.negU, Nb.z;
		"u2", Nb.u, Nb.u, Nb.u;
		"u3", Nb.negU, Nb.u, Nb.negU;
		"u4", Nb.mk1 2, Nb.u, Nb.mk1 2;
		"u5", Nb.mk1 2, Nb.negU, Nb.mk1 (-2);
		"p0", Nb.mk1 4, Nb.mk1 2, Nb.mk1 2;
		"p1", Nb.mk1 3, Nb.mk1 2, Nb.mk 2 3;
		"p2", Nb.mk1 6, Nb.mk1 4, Nb.mk 2 3;
		"n0", Nb.mk1 (-4), Nb.mk1 2, Nb.mk1 (-2);
		"n1", Nb.mk1 4, Nb.mk1 (-2), Nb.mk1 (-2);
		"n2", Nb.mk1 (-4), Nb.mk1 (-2), Nb.mk1 2
	] in
	T.suite "ok" (List.map chk tcs)

(* DM NOTE: divide by zero crashes F# *)
let divTs: T.testT
= T.suite "div" [(*divKoTs; *) divOkTs]

let ts: T.testT
= T.suite "Nb" [cmpTs; cmpzTs; negTs; invTs; addTs; multTs; divTs]
