let x = Var.fromInt 1
let y = Var.fromInt 2
let z = Var.fromInt 3
let a = Var.fromInt 4
let c = Var.fromInt 6
let d = Var.fromInt 5
let e = Var.fromInt 7

let varPr: Var.t -> string
= fun _x ->
	let vars: (Var.t * string) list
	= [x, "x"; y, "y"; z, "z"; a, "a"; c, "c"; d, "d"; e, "e"]
	in
	List.assoc _x vars

let mkc t v c =
	Cstr.mk t (List.map (function (i, x) -> (Nb.mk1 i, x)) v) (Nb.mk1 c)

let eq = mkc Cstr.Eq
let le = mkc Cstr.Le
let lt = mkc Cstr.Lt

let p (l: Cstr.t list): Pol.t =
	match Pol.mk l with
	| None -> failwith "Pol.mk returned None"
	| Some poly -> poly

let eqBnd: Pol.bndT -> Pol.bndT -> bool = fun bnd1 bnd2 ->
	match bnd1, bnd2 with
	| Pol.Infty, Pol.Infty -> true
	| Pol.Open n1, Pol.Open n2
	| Pol.Closed n1, Pol.Closed n2 -> Nb.cmp n1 n2 = 0
	| _, _ -> false

let eqItv: Pol.itvT -> Pol.itvT -> bool = fun itv1 itv2 ->
	eqBnd (Pol.get_low itv1) (Pol.get_low itv2) &&
	eqBnd (Pol.get_up itv1) (Pol.get_up itv2)

(* This is an implementation of full inlining of certificate fragments. *)
module CertInline
= struct
	module IntOrd
	= struct
		type t = int
		let compare = Pervasives.compare
		end

	module Def = Map.Make (IntOrd)

	let pr: Cert.cstr_t Def.t -> string
	=	let pr1: int ->  Cert.cstr_t -> string
		= fun i ce -> Printf.sprintf "%i -> %s\n" i (Cert.prC ce)
		in
		fun d ->
			let d' = Def.mapi pr1 d in
			let l = List.map Pervasives.snd (Def.bindings d') in
			String.concat "\n" l

	let playF: Cert.cstr_t Def.t -> Cert.frag_t -> Cert.frag_t
	= fun d f ->
		let l = List.fold_left
			(fun l (i, n) ->
				if Def.mem i d
				then
					match Def.find i d with
					| Cert.Direct f -> (f, n)::l
					| Cert.SplitEq (f1, f2) ->
						if Nb.cmpz n < 0
						then (f1, n)::l
						else (f2, Nb.neg n)::l
				else ([i, Nb.u], n)::l) [] f
		in
		Cert.mergeF l

	let play1: Cert.cstr_t Def.t -> int * Cert.cstr_t -> int * Cert.cstr_t
	= fun d (i, ce) ->
		match ce with
		| Cert.Direct f -> (i, Cert.Direct (playF d f))
		| Cert.SplitEq (f1, f2) -> (i, Cert.SplitEq (playF d f1, playF d f2))

	let playL: Cert.cstr_t Def.t -> Cert.partialT -> Cert.cstr_t Def.t
	= fun d l ->
		let l' = List.map (fun p -> play1 d p) l in
		List.fold_left (fun d (i, ce) -> Def.add i ce d) Def.empty l'

	let play: Cert.cstr_t Def.t -> Cert.partialT list -> Cert.cstr_t Def.t
	= fun d l -> List.fold_left playL d l

	(* XXX: the Cert.Empty is hackish *)
	let rec playC: Cert.cstr_t Def.t -> Cert.t -> Cert.cstr_t Def.t
	= fun d -> function
		| Cert.Implies l -> playL d l
		| Cert.Empty f -> Def.singleton (-1) (Cert.Direct (playF d f))
		| Cert.Bind (i, _, ce, ce1) ->
			let (i', ce') = play1 d (i, ce) in
			playC (Def.add i' ce' d) ce1

	let equiv: Cert.partialT list -> Cert.t -> Cert.t -> bool
	= fun l ce ce' ->
		let d1 = playC (play Def.empty l) ce in
		let d2 = playC Def.empty ce' in
		Def.equal (fun ce ce' -> Cert.isEqC ce ce') d1 d2
end

module CertChk
= struct

	module IntOrd
	= struct
		type t = int
		let compare = Pervasives.compare
	end

	module CstrSet
	= Set.Make (IntOrd)

	let csmk: int list -> CstrSet.t
	= fun l -> List.fold_left (fun s i -> CstrSet.add i s) CstrSet.empty l

	let polmk: Pol.t -> CstrSet.t
	= fun p -> CstrSet.union
		(csmk (List.map (fun (_, c) -> Cons.get_id c) (Pol.get_eqs p)))
		(csmk (List.map Cons.get_id (Pol.get_ineqs p)))

	(** Check that bindings don't define constraints with an already allocated
	identifier. *)
	let rec redefFree: CstrSet.t -> Cert.t -> bool
	= fun s -> function
		| Cert.Implies _ -> true
		| Cert.Empty _ -> true
		| Cert.Bind (i, _, _, ce) ->
			if CstrSet.mem i s
			then false
			else redefFree (CstrSet.add i s) ce

	module Def
	= Map.Make (IntOrd)

	type defMapT
	= Cert.cstr_t Def.t

	let dmMk: Cert.t -> defMapT * Cert.t
	=	let rec dig: defMapT -> Cert.t -> defMapT * Cert.t
		= fun dm -> function
			| Cert.Implies _ as ce -> (dm, ce)
			| Cert.Empty _ as ce -> (dm, ce)
			| Cert.Bind (i, _, ce, ce') -> dig (Def.add i ce dm) ce'
		in
		fun ce -> dig Def.empty ce

	let rec fragRewind: defMapT -> Cert.frag_t -> Cert.frag_t
	=	let pickDef: defMapT -> int * Nb.t -> Cert.frag_t * Nb.t
		= fun dm (i, n) ->
			if Def.mem i dm
			then
				match Def.find i dm with
				| Cert.Direct f -> (f, n)
				| Cert.SplitEq (f, f') ->
					if Nb.cmpz n < 0
					then (f, n)
					else (f', Nb.neg n)
			else (Cert.mkF [i, Nb.u], n)
		in
		fun dm f ->
			let f' = Cert.mergeF (List.map (pickDef dm) f) in
			if Cert.isEqF f f'
			then f'
			else fragRewind dm f'

	let ineqRewind: defMapT -> Cert.cstr_t -> Cert.cstr_t option
	= fun dm -> function
		| Cert.Direct f -> Some (Cert.Direct (fragRewind dm f))
		| Cert.SplitEq (f, f') -> None

	let subst
	= fun dm (i, n) ->
		if Def.mem i dm
		then (Def.find i dm, n)
		else (Cert.Direct (Cert.mkF [i, Nb.u]), n)

	let isSplit
	= fun (ce, _) ->
		match ce with
		| Cert.Direct _ -> false
		| Cert.SplitEq (_, _) -> true

	let substD
	= fun (ce, n) ->
		match ce with
		| Cert.Direct f -> (f, n)
		| Cert.SplitEq (_, _) -> assert false

	let substS1
	= fun (ce, n) ->
		match ce with
		| Cert.Direct _ -> assert false
		| Cert.SplitEq (f, f') ->
			if Nb.cmpz n < 0
			then (f, n)
			else (f', Nb.neg n)

	let substS2
	= fun (ce, n) ->
		match ce with
		| Cert.Direct _ -> assert false
		| Cert.SplitEq (f', f) ->
			if Nb.cmpz n < 0
			then (f, n)
			else (f', Nb.neg n)

	let rec eqRewind: defMapT -> Cert.cstr_t -> Cert.cstr_t option
	= fun dm -> function
		| Cert.Direct f ->
			let f' = List.map (subst dm) f in
			let (f1, f2) = List.partition isSplit f' in
			if List.length f1 = 0
			then
				let f2' = Cert.mergeF (List.map substD f2) in
				if Cert.isEqF f f2'
				then Some (Cert.Direct f2')
				else eqRewind dm (Cert.Direct f2')
			else
				let f2' = List.map substD f2 in
				let f2'opp = List.map (fun (f, n) -> (f, Nb.neg n)) f2' in
				let fL = Cert.mergeF ((List.map substS1 f1) @ f2') in
				let fR = Cert.mergeF ((List.map substS2 f1) @ f2'opp) in
				eqRewind dm (Cert.SplitEq (fL, fR))
		| Cert.SplitEq (fL, fR) ->
			match ineqRewind dm (Cert.Direct fL) with
			| Some (Cert.SplitEq (_, _)) -> None
			| None -> None
			| Some (Cert.Direct fL') ->
				match ineqRewind dm (Cert.Direct fR) with
				| Some (Cert.Direct fR') -> Some (Cert.SplitEq (fL', fR'))
				| Some (Cert.SplitEq (_, _)) -> None
				| None -> None

	let rewindAux: CstrSet.t -> Cert.t -> Cert.t option
	= fun eqs ce ->
		let (dm, ce') = dmMk ce in
		match ce' with
		| Cert.Empty f -> Some (Cert.Empty (fragRewind dm f))
		| Cert.Bind (_, _, _, _) -> assert false
		| Cert.Implies l ->
			let (l1, l2) = List.partition (fun (i, _) -> CstrSet.mem i eqs) l in
			let rw f
			= function
				| None -> fun _ -> None
				| Some l -> fun (i, ce) ->
					match f ce with
					| None -> None
					| Some ce' -> Some ((i, ce')::l)
			in
			let ml1' = List.fold_left (rw (eqRewind dm)) (Some []) l1 in
			let ml2' = List.fold_left (rw (ineqRewind dm)) (Some []) l2 in
			match ml1', ml2' with
			| Some l1', Some l2' -> Some (Cert.Implies (l1' @ l2'))
			| None, Some _
			| Some _, None
			| None, None -> None

	let rewind: Pol.t option -> Cert.t -> Cert.t option
	= fun maybeP ce ->
		match maybeP with
		| Some p ->
			let eqIds = List.map (fun (_, c) -> Cons.get_id c) (Pol.get_eqs p) in
			rewindAux (csmk eqIds) ce
		| None -> rewindAux (CstrSet.empty) ce

end

(* Pol.certCombineFwd *)
let certCombineFwdTs: T.testT
=	let chk (nm, l, ce0)
	= fun st ->
		let ce = Pol.certCombineFwd l ce0 in
		if CertInline.equiv l ce0 ce
		then T.succeed st
		else
			let estr
			= Printf.sprintf "l <> ce\nce:\n%s\ndl:\n%s\ndce:\n%s\n"
				(Cert.pr ce)
				(CertInline.pr (CertInline.play CertInline.Def.empty l))
				(CertInline.pr (CertInline.playC CertInline.Def.empty ce))
			in
			T.fail nm estr st
	in
	let tcs: (string * (int * Cert.cstr_t) list list * Cert.t) list
	= [
		"direct0", [], Cert.Implies [1, Cert.Direct (Cert.mkF [1, Nb.u])];
		"direct1", [[]], Cert.Implies [1, Cert.Direct (Cert.mkF [1, Nb.u])];

		"simpl0", [[1, Cert.Direct (Cert.mkF [1, Nb.u])]],
			Cert.Implies [1, Cert.Direct (Cert.mkF [1, Nb.u])];

		"simpl1", [[1, Cert.SplitEq (Cert.mkF [1, Nb.u], Cert.mkF [2, Nb.u])]],
			Cert.Implies [1, Cert.Direct (Cert.mkF [1, Nb.negU])];

		"simpl2", [[
			1, Cert.Direct (Cert.mkF [1, Nb.u]);
			2, Cert.Direct (Cert.mkF [1, Nb.mk1 2])
		]], Cert.Empty (Cert.mkF [1, Nb.mk1 2; 2, Nb.u]);

		(* non-unit coefficient *)
		"two0", [[1, Cert.Direct (Cert.mkF [1, Nb.mk1 2])]],
			Cert.Implies [1, Cert.Direct (Cert.mkF [1, Nb.u])];

		(* non-overlapping index sets *)
		"two1", [[0, Cert.Direct (Cert.mkF [1, Nb.mk1 2])]],
			Cert.Implies [1, Cert.Direct (Cert.mkF [0, Nb.u])];

		(* linear combination *)
		"two2", [[
			1, Cert.Direct (Cert.mkF [1, Nb.mk1 2; 2, Nb.mk1 3]);
			2, Cert.Direct (Cert.mkF [2, Nb.u; 4, Nb.mk1 3])
		]], Cert.Implies [1, Cert.Direct (Cert.mkF [1, Nb.u; 2, Nb.u])];

		(* three steps *)
		"three0", [[
			1, Cert.Direct (Cert.mkF [1, Nb.u; 2, Nb.u]);
			2, Cert.Direct (Cert.mkF [1, Nb.mk1 2; 3, Nb.negU])
		]; [
			1, Cert.Direct (Cert.mkF [1, Nb.u]);
			2, Cert.Direct (Cert.mkF [1, Nb.u; 2, Nb.negU])
		]], Cert.Implies [1, Cert.Direct (Cert.mkF [1, Nb.u; 2, Nb.u])];
	] in
	T.suite "certCombineFwd" (List.map chk tcs)

(* Pol.add *)
let addTcs =
	let p': (Var.t * int * Cstr.t) list -> (int * Cstr.t) list -> Pol.t
	= fun eqs ineqs -> {
		Pol.nxt = 1000;
		Pol.eqs = List.map (fun (x, i, c) -> (x, Cons.mk i c)) eqs;
		Pol.ineqs = List.map (fun (i, c) -> Cons.mk i c) ineqs}
	in [
	"noEqs0", false, le [1, x] 0, p [
		le [1, y] 0 ],
		Pol.Added (p' [] [
			0, le [1, y] 0;
			1, le [1, x] 0],
			(Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [1, Nb.u])],
			Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [1, Nb.u])]));

	"eqs0", false, le [1, x; 1, y] 0, p [
		eq [1, x; -1, y] 0 ],
		Pol.Added (p' [
			x, 0, eq [1, x; -1, y] 0]
			[1, le [1, y] 0],
			(Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [0, Nb.mk 2 (-1); 1, Nb.mk 2 1])],
			Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.mk1 2])]));

	"eqs1", false, le [1, x; 1, y] 0, p [
		eq [1, x; -1, z] 0],
		Pol.Added (p' [
			x, 0, eq [1, x; -1, z] 0]
			[1, le [1, y; 1, z] 0],
			(Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [0, Nb.negU; 1, Nb.u])],
			Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.u])]));

	"eqs2", false, eq [1, x; 1, y] 1, p [
		eq [1, x] 0],
		Pol.Added (p' [
			x, 0, eq [1, x] 0;
			y, 1, eq [1, y] 1] [],
			(Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [0, Nb.negU; 1, Nb.u])],
			Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.u])]));

	"eqs3", false, eq [1, x; 1, y] 1, p [
		eq [1, x] 0;
		le [1, y; 1, z] 0],
		Pol.Added (p' [
			x, 0, eq [1, x] 0;
			y, 2, eq [1, y] 1]
			[1, le [1, z] (-1)],
			(Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.u; 2, Nb.negU]);
				2, Cert.Direct (Cert.mkF [0, Nb.negU; 2, Nb.u])],
			Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [1, Nb.u; 2, Nb.u]);
				2, Cert.Direct (Cert.mkF [0, Nb.u; 2, Nb.u])]));

(* When the inequality set is rewritten, the two inequalities become
syntactically redundant. *)
	"eqs4", false, eq [1, x] 0, p [
		le [1, x; 1, y] 1;
		le [1, y] 1],
		Pol.Added (p'
			[x, 2, eq [1, x] 0]
			[1, le [1, y] 1],
			(Cert.Implies [
				1, Cert.Direct (Cert.mkF [1, Nb.u]);
				2, Cert.Direct (Cert.mkF [2, Nb.u])],
			Cert.Implies [
				0, Cert.Direct (Cert.mkF [1, Nb.u; 2, Nb.u]);
				1, Cert.Direct (Cert.mkF [1, Nb.u]);
				2, Cert.Direct (Cert.mkF [2, Nb.u])]));

	"newEq0", false, eq [1, x] 0, p [
		le [1, x; 1, y] 2 ],
		Pol.Added (p' [
			x, 1, eq [1, x] 0]
			[0, le [1, y] 2],
			(Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.negU]);
				1, Cert.Direct (Cert.mkF [1, Nb.u])],
			Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u; 1, Nb.u]);
				1, Cert.Direct (Cert.mkF [1, Nb.u])]));

	"implicitEq0", false, le [1, x] 0, p [
		le [1, y] 0;
		le [-1, x] 0 ],
		Pol.Added (p' [
			x, 4, eq [1, x] 0]
			[0, le [1, y] 0],
			(Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				4, Cert.SplitEq (Cert.mkF [2, Nb.u], Cert.mkF [1, Nb.u])],
			Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [4, Nb.negU]);
				2, Cert.Direct (Cert.mkF [4, Nb.u])]));

	"implicitEqs0", false, le [1, z; -1, x] 0, p [
		le [1, y; -1, z] 0;
		le [1, x; -1, y] 0 ],
		Pol.Added (p' [
			y, 4, eq [1, y; -1, z] 0;
			x, 5, eq [1, x; -1, y] 0] [],
			(Cert.Implies [
				5, Cert.SplitEq (
					Cert.mkF [1, Nb.u],
					Cert.mkF [0, Nb.u; 2, Nb.u]);
				4, Cert.SplitEq (
					Cert.mkF [0, Nb.u],
					Cert.mkF [1, Nb.u; 2, Nb.u])],
			Cert.Implies [
				0, Cert.Direct (Cert.mkF [4, Nb.u]);
				1, Cert.Direct (Cert.mkF [5, Nb.u]);
				2, Cert.Direct (Cert.mkF [4, Nb.negU; 5, Nb.negU])]));

	"implicitEqs1", false, le [1, z; -1, x] 0, p [
		le [1, y; -1, z] 0;
		le [1, x; -1, y] 0;
		le [1, x; 1, a] 1 ],
		Pol.Added (p'
			[x, 5, eq [1, x; -1, z] 0;
			y, 6, eq [1, y; -1, z] 0]
			[2, le [1, z; 1, a] 1],
			(Cert.Implies [
				2, Cert.Direct
					(Cert.mkF [0, Nb.u; 1, Nb.u; 2, Nb.u; 3, Nb.mk1 2]);
				5, Cert.SplitEq (
					Cert.mkF [0, Nb.u; 1, Nb.u],
					Cert.mkF [0, Nb.u; 1, Nb.u; 3, Nb.mk1 2]);
				6, Cert.SplitEq (
					Cert.mkF [0, Nb.u],
					Cert.mkF [1, Nb.u; 3, Nb.u])],
			Cert.Implies [
				0, Cert.Direct (Cert.mkF [6, Nb.u]);
				1, Cert.Direct (Cert.mkF [5, Nb.u; 6, Nb.negU]);
				2, Cert.Direct (Cert.mkF [2, Nb.u; 5, Nb.u]);
				3, Cert.Direct (Cert.mkF [5, Nb.negU])]));

	(* substitutions in the building blocks of an implicit equality *)
	"implicitEqs2", false, le [1, x; 1, y] 0, p [
		eq [1, y] 0;
		le [-1, x] 0 ],
		Pol.Added (p' [
			x, 4, eq [1, x] 0;
			y, 0, eq [1, y] 0] [],
			(Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				4, Cert.SplitEq (
					Cert.mkF [0, Nb.negU; 2, Nb.u],
					Cert.mkF [1, Nb.u])],
			Cert.Implies [
				0, Cert.Direct (Cert.mkF [0, Nb.u]);
				1, Cert.Direct (Cert.mkF [4, Nb.negU]);
				2, Cert.Direct (Cert.mkF [0, Nb.u; 4, Nb.u])]));

	"emptyEq0", true, eq [-1, z] 2, p [
		eq [-1, x] 2;
		eq [-1, y] 0;
		le [-1, z] (-6) ],
		Pol.Contrad (Cert.Empty (Cert.mkF [2, Nb.u; 3, Nb.negU]));

	"emptyLe0", true, le [1, x] 0, p [
		le [-1, x] (-1) ],
		Pol.Contrad (Cert.Empty (Cert.mkF [0, Nb.u; 1, Nb.u]));

	"emptyLe1", true, le [1, x; 1, y] 0, p [
		le [-1, x] (-1);
		eq [1, y] 0 ],
		Pol.Contrad (Cert.Empty (Cert.mkF [0, Nb.u; 1, Nb.negU; 2, Nb.u]));

	"emptyLe2", true, le [-1, x] (-3), p [
		eq [1, x] 2;
		eq [1, y] 1],
		Pol.Contrad (Cert.Empty (Cert.mkF [0, Nb.u; 2, Nb.u]))
]

let addTs: T.testT
=
	let chkBot (t, empty, c, p, _) = fun state ->
		match Pol.add p c with
		| Pol.Added (_, _) ->
			if not empty then
				T.succeed state
			else
				T.fail t "should be bottom" state
		| Pol.Contrad _ ->
			if empty then
				T.succeed state
			else
				T.fail t "should not be bottom" state
	in
	let chk_subst (t, isEmpty, c, p, _) = fun state ->
		match Pol.add p c, isEmpty with
		| Pol.Contrad _, true
		| Pol.Contrad _, false
		| Pol.Added (_, _), true -> T.succeed state
		| Pol.Added (p1, _), false ->

		let defs = List.map (fun (x, _) -> x) (Pol.get_eqs p1) in
		let chk1 c = List.for_all
			(fun x -> Nb.cmpz (Vec.get (Cstr.get_v (Cons.get_c c)) x) = 0) defs
		in
		if List.for_all chk1 (Pol.get_ineqs p1) then
			T.succeed state
		else
			let err = "defined variables in inequality set\n" ^ (Pol.prExt varPr p1) in
			T.fail t err state
	in
	let chk_inv (nm, _, c, p, _)
	= fun st ->
		match Pol.add p c with
		| Pol.Contrad _ -> T.succeed st
		| Pol.Added (p', _) ->
			match Pol.invChk p' with
			| true, _ -> T.succeed st
			| false, e -> T.fail nm e st
	in
(** Check the well-formedness of the emitted certificates. *)
	let chkCertSyn (nm, _, c, p, _)
	= fun st ->
		let cs = CertChk.polmk p in
		match Pol.add p c with
		| Pol.Added (p', (fwd, bwd)) ->
			let cs' = CertChk.polmk p' in
			if CertChk.redefFree cs fwd && CertChk.redefFree cs' bwd
			then T.succeed st
			else
				let e = "badly bound certificate" in
				T.fail nm e st
		| Pol.Contrad ce ->
			if CertChk.redefFree cs ce
			then T.succeed st
			else
				let e = "badly bound certificate" in
				T.fail nm e st
	in
	let chkFwdCert (nm, _, c, p, r)
	=	let handleContrad: Cert.t -> Cert.t -> T.stateT -> T.stateT
		= fun ce ce' st ->
			match CertChk.rewind None ce, CertChk.rewind None ce' with
			| None, _ | _, None -> T.fail nm "bad certificate" st
			| Some ce2, Some ce2' ->
				if Cert.isEq ce2 ce2'
				then T.succeed st
				else
					let e = "bad forward certificate" in
					T.fail nm e st
		in
		let handleAdded
		= fun (ap, (afwd, _)) (ep, (efwd, _)) st ->
			match CertChk.rewind (Some ep) afwd, CertChk.rewind (Some ep) efwd with
			| None, _ | _, None -> T.fail nm "bad certificate" st
			| Some a', Some e' ->
				if Cert.isEq a' e'
				then T.succeed st
				else
					let e = Printf.sprintf "bad certificate" in
					T.fail nm e st
		in
		fun st ->
			(match Pol.add p c, r with
			| Pol.Added (p, ce), Pol.Added (q, ce') -> handleAdded (p, ce) (q, ce')
			| Pol.Contrad ce, Pol.Contrad ce' -> handleContrad ce ce'
			| Pol.Added (_, _), Pol.Contrad _
			| Pol.Contrad _, Pol.Added (_, _) -> T.succeed) st
	in
	let chkBwdCert (nm, _, c, p, r)
	= fun st ->
		match Pol.add p c, r with
		| Pol.Contrad _, Pol.Contrad _
		| Pol.Added (_, _), Pol.Contrad _
		| Pol.Contrad _, Pol.Added (_, _) -> T.succeed st
		| Pol.Added (ap, (_, ce)), Pol.Added (ep, (_, ce')) ->
			match CertChk.rewind (Some p) ce, CertChk.rewind (Some p) ce' with
			| None, _ | _, None -> T.fail nm "bad certificate" st
			| Some ce1', Some ce2' ->
				if Cert.isEq ce1' ce2'
				then T.succeed st
				else
					let e = Printf.sprintf
	"incorrect certificate\nce1':\n%s\nce2':\n%s\nap:\n%s\nep:\n%s\np:\n%s\nc: %s\n"
						(Cert.pr ce1') (Cert.pr ce2')
						(Pol.prExt varPr ap) (Pol.prExt varPr ep)
						(Pol.prExt varPr p) (Cstr.pr varPr c)
					in
					T.fail nm e st
	in
	T.suite "add" [
		T.suite "bot" (List.map chkBot addTcs);
		T.suite "subst" (List.map chk_subst addTcs);
		T.suite "inv" (List.map chk_inv addTcs);
		T.suite "cert_syntax" (List.map chkCertSyn addTcs);
		T.suite "forward_cert" (List.map chkFwdCert addTcs);
		T.suite "backward_cert" (List.map chkBwdCert addTcs)
	]

(* Pol.meet *)
let meetTcs: (string * Pol.t * Pol.t * Pol.t option) list
= [
	"trivial0", p [], p [], Some (p []);

(* simple preservation of the constraints in the inputs *)
	"trivial1", p [
		le [1, x] 0
	], p [], Some (p [
		le [1, x] 0
	]);

	"trivial2", p [], p [
		le [1, x] 0
	], Some (p [
		le [1, x] 0
	]);

	"trivial3", p [
		eq [1, x] 0
	], p [], Some (p [
		eq [1, x] 0
	]);

	"trivial4", p [], p [
		eq [1, x] 0
	], Some (p [
		eq [1, x] 0
	]);

	"trivial5", p [
		le [1, x] 0
	], p [
		le [1, y] 0
	], Some (p [
		le [1, x] 0;
		le [1, y] 0
	]);

	"trivial6", p [
		eq [1, x] 0
	], p [
		le [1, y] 0
	], Some (p [
		eq [1, x] 0;
		le [1, y] 0
	]);

	"trivial7", p [
		le [1, x] 0
	], p [
		eq [1, y] 0
	], Some (p [
		le [1, x] 0;
		eq [1, y] 0
	]);

	"trivial6", p [
		eq [1, x] 0
	], p [
		eq [1, y] 0
	], Some (p [
		eq [1, x] 0;
		eq [1, y] 0
	]);

(* redundant constraint elimination *)
	"red0", p [
		le [1, x] 0
	], p [
		le [1, x] 1
	], Some (p [
		le [1, x] 0
	]);

	"red1", p [
		le [1, x] 0;
		le [1, y] 1
	], p [
		le [1, x; 1, y] 1
	], Some (p [
		le [1, x] 0;
		le [1, y] 1
	]);

	"red2", p [
		le [1, x] 1
	], p [
		le [1, x] 0
	], Some (p [
		le [1, x] 0
	]);

	"red3", p [
		le [1, x; 1, y] 1
	], p [
		le [1, x] 0;
		le [1, y] 1
	], Some (p [
		le [1, x] 0;
		le [1, y] 1
	]);

	"red4",
		p [eq [1, x] 0],
		p [le [1, x] 0],
		Some (p [eq [1, x] 0]);

(* variable substitutions using equalities *)
	"eq0", p [
		eq [1, x] 0
	], p [
		le [1, x; 1, y] 1
	], Some (p [
		eq [1, x] 0;
		le [1, y] 1
	]);

	"eq1", p [
		eq [1, x] 0
	], p [
		le [1, x] 1
	], Some (p [
		eq [1, x] 0
	]);

	"eq2", p [
		eq [1, x] 0;
		le [1, y; 1, z] 1
	], p [
		eq [1, y] 2;
		le [1, x; 1, z] 1
	], Some (p [
		eq [1, x] 0;
		eq [1, y] 2;
		le [1, z] (-1)
	]);

(* new implicit equalities *)
	"implicit0", p [
		le [1, x] 0
	], p [
		le [-1, x] 0
	], Some (p [
		eq [1, x] 0
	]);

	"implicit1", p [
		le [1, x] 0
	], p [
		le [-1, x] 0;
		le [1, x; 1, z] 2
	], Some (p [
		eq [1, x] 0;
		le [1, z] 2
	]);

	"implicit2", p [
		le [1, x] 0;
		eq [1, y] 2
	], p [
		le [-1, x] 0;
		le [1, x; 1, y; 1, z] 2
	], Some (p [
		eq [1, y] 2;
		eq [1, x] 0;
		le [1, z] 0
	]);

(* empty polyhedra *)
	"empty0", p [
		le [1, x] 0
	], p [
		le [-1, x] (-2)
	], None;

	"empty1", p [
		eq [1, x] 0
	], p [
		le [-1, x] (-2)
	], None;

	"empty2", p [
		le [1, x] 0
	], p [
		eq [-1, x] (-2)
	], None;

	"empty3", p [
		eq [1, x] 0
	], p [
		eq [1, x] (-2)
	], None;

	"empty4", p [
		le [1, x; 1, y] 1;
		le [-1, y] (-2)
	], p [
		le [-1, x] 0
	], None;

	"empty5", p [
		le [1, x; 1, y] 1
	], p [
		eq [1, x] 0;
		le [-1, y] (-2)
	], None
]

let meetTs: T.testT
=	let chk (nm, p1, p2, maybeP)
	= fun st ->
		let noneExpected: Pol.meetT -> T.stateT
		= function
			| Pol.Contrad _ -> T.succeed st
			| Pol.Added (_, _) ->
				let estr = "expected None but got Some _" in
				T.fail nm estr st
		in
		let someExpected: Pol.t -> Pol.meetT -> T.stateT
		= fun p -> function
			| Pol.Contrad _ ->
				let estr = "expected Some _ but got None" in
				T.fail nm estr st
			| Pol.Added (p', _) ->
				if Pol.isEqSyn p p'
				then T.succeed st
				else
					let estr = Printf.sprintf "expected\n%s\ngot\n%s\n"
						(Pol.prExt varPr p) (Pol.prExt varPr p')
					in
					T.fail nm estr st
		in
		let res = Pol.meet p1 p2 in
		match maybeP with
		| None -> noneExpected res
		| Some p -> someExpected p res
	in
	T.suite "meet" (List.map chk meetTcs)

(* Pol.isIncl *)
let isInclTcs = [
	"simple0", true, p [
		le [1, x; 1, y] 2
	], p [
		le [1, x; 1, y] 3 ];

	"simple1", true, p [
		le [1, x] 2
	], p [
		le [1, x] 3 ];

	"simple2", true, p [
		le [1, x] 2
	], p [
		le [2, x] 5 ];

	"eq0", true, p [
		le [1, x] 1;
		le [-1, x] (-1)
	], p [
		eq [1, x] 1 ];

	"fail0", false, p [
		le [1, x] 2
	], p [
		le [1, x] 1 ];

	"eq1", true, p [
		eq [1, x] 0
	], p [
		eq [1, x] 0 ];

	"eq2", true, p [
		eq [-1, x] 0
	], p [
		eq [-1, x] 0 ];

	"decompress", false, p [
		le [-1, y] (-1);
		le [-1, x; 16, y] 15;
		le [-1, x] (-17)
	], p [
		le [1, x; -16, y] 17 ];

	"sylvain", false, p [
		le [2, x] 3
	], p [
		le [1, x] 1 ];

	"splx0", true, p [
		le [1, x] 2;
		le [2, y] 1;
	], p [
		le [1, x; 2, y] 3 ];

	"order", true, p [
		le [1, x] 2;
		le [1, y] 3
	], p [
		le [1, y] 3;
		le [1, x] 2];

	"branch", true, {
	Pol.nxt = 3;
	Pol.eqs = [
		(x, {Cons.id = 2; Cons.c = eq [1, x] 3; Cons.f = [0, Nb.u; 2, Nb.u]})];
	Pol.ineqs = []
	}, {
	Pol.nxt = 4;
	Pol.eqs = [];
	Pol.ineqs = [
		{Cons.id = 3; Cons.c = le [-1, x] 0; Cons.f = [3, Nb.u]};
		{Cons.id = 2; Cons.c = le [1, x] 3; Cons.f = [2, Nb.u]}]}
]

(* XXX: add syntactic check on the certificates *)
let isInclTs: T.testT
=
	let chk_res (name, r, p1, p2) = fun state ->
		let (ok1, e1) = Pol.invChk p1 in
		if not ok1
		then T.fail name ("bad p1:\n" ^ e1) state
		else
			let (ok2, e2) = Pol.invChk p2 in
			if not ok2
			then T.fail name ("bad p2:\n" ^ e2) state
			else
				let dump: Pol.t -> Pol.t -> string
				= fun p1 p2 ->
					Printf.sprintf "p1:\n%s\np2:\n%s"
						(Pol.prExt varPr p1) (Pol.prExt varPr p2)
				in
				match Pol.isIncl p1 p2 with
				| Pol.NoIncl ->
					if r
					then T.fail name ("no inclusion\n" ^ (dump p1 p2)) state
					else T.succeed state
				| Pol.Incl _ ->
					if r
					then T.succeed state
					else T.fail name ("unexpected inclusion\n" ^ (dump p1 p2)) state
	in
	T.suite "isIncl" (List.map chk_res isInclTcs)

(* Pol.project *)
let projectTcs = [
	"empty", x, p [], p [];
	"no_x0", x, p [eq [1, y] 0], p [eq [1, y] 0];
	"no_x1", x, p [eq [1, y] 0; eq [2, y; 1, z] 0],
		p [eq [1, y] 0; eq [2, y; 1, z] 0];
	"only_x", x, p [eq [1, x] 0], p [];
	"simple0", x, p [le [1, x; 2, y] 1; le [1, y] 2], p [le [1, y] 2];
	"subst0", x, p [eq [1, x; -1, y] 0; le [1, x] 1], p [le [1, y] 1];
	"fme0", x, p [le [1, x] 0; le [-1, x; 1, y] 1], p [le [1, y] 1];
	"trivial0", x, p [le [1, x; 1, y] 0; le [-1, x; -1, y] 1], p [];
]

(* XXX: Add syntactic criteria for the certificates *)
let projectTs: T.testT
=
	let chk_res (t, v, p, r) = fun state ->
		let (p1, _) = Pol.project p v in
		if Pol.isEqSyn p1 r then
			T.succeed state
		else
			let estr =
				"projecting x from polyhedron P:\n" ^ (Pol.prExt varPr p) ^
				"\ngives polyhedron P1:\n" ^ (Pol.prExt varPr p1) ^
				"\nwhich is not syntactically equal to the expected R:\n" ^
					(Pol.prExt varPr r)
			in
			T.fail t estr state
	in
	T.suite "project" (List.map chk_res projectTcs)

(* Pol.projectM *)
let projectMTcs = [
	"empty", [x], p [], p [];
	"no_x0", [x], p [
		eq [1, y] 0
	], p [
		eq [1, y] 0 ];

	"no_x1", [x], p [
		eq [1, y] 0;
		eq [2, y; 1, z] 0
	], p [
		eq [1, y] 0;
		eq [2, y; 1, z] 0 ];

	"only_x", [x], p [
		eq [1, x] 0
	], p [];

	"simple0", [x], p [
		le [1, x; 2, y] 1;
		le [1, y] 2
	], p [
		le [1, y] 2 ];

	"subst0", [x], p [
		eq [1, x; -1, y] 0;
		le [1, x] 1
	], p [
		le [1, y] 1 ];

	"fme0", [x], p [
		le [1, x] 0;
		le [-1, x; 1, y] 1
	], p [
		le [1, y] 1 ];

	"trivial0", [x], p [
		le [1, x; 1, y] 0;
		le [-1, x; -1, y] 1
	], p [];

	"le0", [x; z], p [
		le [1, x; 2, y; -3, z] 0;
		le [-2, x; -1, z] 0;
		le [3, y; 1, z] 0
	], p [
		le [1, y] 0 ];

	"eq0", [x; z], p [
		le [1, x; 2, y; -3, z] 0;
		le [-2, x; -1, z] 0;
		eq [1, y; -1, z] 0
	], p [
		le [-1, y] 0 ]
]

let projectMTs: T.testT
=
	let chk_res (t, l, p, r) = fun state ->
		let (p1, _) = Pol.projectM p l in
		if Pol.isEqSyn p1 r then
			T.succeed state
		else
			let estr =
				"projecting" ^
				(List.fold_left (fun s x -> s ^ " " ^ (varPr x)) "" l) ^
 				" from polyhedron P:\n" ^ (Pol.prExt varPr p) ^
				"\ngives polyhedron P1:\n" ^ (Pol.prExt varPr p1) ^
				"\nwhich is not syntactically equal to the expected R:\n" ^
					(Pol.prExt varPr r)
			in
			T.fail t estr state
	in
	T.suite "projectM" (List.map chk_res projectMTcs)

(* Pol.joinSetup *)
let joinSetupTs: T.testT
=
	let x = Var.fromInt 1 in
	let y = Var.fromInt 2 in
	let alpha = Var.fromInt 3 in
	let x1 = Var.fromInt 4 in
	let y1 = Var.fromInt 6 in
	let chk_pol (t, _, p1, p2, p0) = fun state ->
		let (_, _, p, _) = Pol.joinSetup (Var.next y) p1 p2 in
		if Pol.isEqSyn p0 p then
			T.succeed state
		else
			let estr =
				"expected p:\n" ^ (Pol.prExt varPr p0) ^
				"\n\nactual p:\n" ^ (Pol.prExt varPr p)
			in
			T.fail t estr state
	in
	let chk_vars (t, l0, p1, p2, _) = fun state ->
		let (_, _, p, l) = Pol.joinSetup (Var.next y) p1 p2 in
		let incl l1 l2 =
			List.for_all (fun x -> List.exists ((=) x) l2) l1
		in
		if incl l l0 && incl l0 l then
			T.succeed state
		else
			T.fail t "bad var list" state
	in
	let chk_frag (t, _, p1, p2, _) = fun state ->
		let (_, _, p, _) = Pol.joinSetup (Var.next y) p1 p2 in
		let initP = { p with
			Pol.eqs = EqSet.initc (Pol.get_eqs p);
			Pol.ineqs = IneqSet.initc (Pol.get_ineqs p) }
		in
		let chk =
			List.for_all2 (fun c1 c2 ->
				List.for_all2 (fun (x1, n1) (x2, n2) ->
					x1 = x2 && Nb.cmp n1 n2 = 0) (Cons.get_f c1) (Cons.get_f c2))
		in
		if chk (EqSet.list (Pol.get_eqs p)) (EqSet.list (Pol.get_eqs initP)) &&
			chk (IneqSet.list (Pol.get_ineqs p)) (IneqSet.list (Pol.get_ineqs initP)) then
			T.succeed state
		else
			let e = "not equal\np:\n" ^ (Pol.prExt varPr p) ^ "\ninitP:\n" ^ (Pol.prExt varPr initP) in
			T.fail t e state
	in
	let tcs = [
		"one", [alpha; x1; y1], p [
			eq [1, x] 1;
			le [-1, y] 1
		], p [
			eq [1, x] 1;
			le [1, y] 2
		], p [
			eq [1, x; -1, x1; 1, alpha] 1;
			eq [1, x1; -1, alpha] 0;
			le [-1, y1; -1, alpha] 0;
			le [1, y; -1, y1; 2, alpha] 2;
			le [-1, alpha] 0;
			le [1, alpha] 1 ];

		"all", [alpha; x1; y1], p [
			eq [1, x] 1
		], p [
			eq [1, y] 1
		], p [
			eq [1, y; -1, y1; 1, alpha] 1;
			eq [1, x1; -1, alpha] 0;
			le [-1, alpha] 0;
			le [1, alpha] 1 ]
	] in
	T.suite "joinSetup" [
		T.suite "pol" (List.map chk_pol tcs);
		T.suite "vars" (List.map chk_vars tcs);
		T.suite "frag" (List.map chk_frag tcs)
	]

(* Pol.join *)
let joinTcs: (string * Pol.t * Pol.t * Pol.t) list
= [
	"one", p [
		eq [1, x] 1;
		le [-1, y] 1
	], p [
		eq [1, x] 1;
		le [1, y] 2
	], p [
		eq [1, x] 1
	];

	"all",p [
		eq [1, x] 1
	], p [
		eq [1, y] 1
	], p [];

	"chull969", p [
		eq [1, c; -1, d] (-4080);
		eq [-1, e] (-15);
		le [1, z] 6;
		le [1, c] 15;
		le [-1, a] (-256);
		le [-1, z] (-2);
		le [-1, y] (-1);
		le [-1, x; 1, z] 0
	], p [
		le [-16, e; 1, d] 3855
	], p [
		le [1, d; -16, e] 3855
	];

	"sylvain", p [
		eq [-1, x; 1, y] 5;
		le [1, x] 5;
		le [-1, x] (-1)
	], p [
		eq [1, x; -1, y] 5;
		le [1, x] 10;
		le [-1, x] (-6)
	], p [
		le [1, x; -1, y] 5;
		le [1, x; 1, y] 15;
		le [-1, x; -1, y] (-7);
		le [-1, x; 1, y] 5
	];

	"jdphuff_chull777", p [
		eq [-1, x] 0;
		le [-1, y] 0;
		le [-1, z; 1, a] 1
	], p [
		eq [-1, x] 0;
		le [-1, y] 0;
		eq [-1, z; 1, a] 1
	], p [
		eq [1, x] 0;
		le [-1, y] 0;
		le [1, a; -1, z] 1
	];

	"notClosed0", p [
		lt [1, y] 3;
	], p [
		le [1, x; 1, y] 5;
		le [-1, x; 1, y] 1
	], p [
		le [1, y] 3
	];

	"dots0", p [
		eq [1, x] 0;
		eq [1, y] 0
	], p [
		eq [1, x] 1;
		eq [1, y] 2
	], p [
		eq [2, x; -1, y] 0;
		le [1, y] 2;
		le [-1, y] 0
	];

	"chk_guess0", p [
		eq [1, y] 0
	], p [
		le [-1, y] (-1)
	], p [
		le [-1, y] 0
	]
]

(* XXX: certificates should be syntactically checked *)
let joinTs: T.testT
=
	let chk1 (name, p1, p2, _) = fun state ->
		let (p, _) = Pol.join p1 p2 in
		match Pol.isIncl p1 p with
		| Pol.Incl _ -> T.succeed state
		| Pol.NoIncl ->
			let err = Printf.sprintf "p1 not in p\np:\n%s" (Pol.prExt varPr p) in
			T.fail name err state
	in
	let chk2 (name, p1, p2, _) = fun state ->
		let (p, _) = Pol.join p1 p2 in
		match Pol.isIncl p2 p with
		| Pol.Incl _ -> T.succeed state
		| Pol.NoIncl ->
			let err = Printf.sprintf "p2 not in p\np:\n%s" (Pol.prExt varPr p) in
			T.fail name err state
	in
	let chkR (name, p1, p2, pR) = fun state ->
		let (p, _) = Pol.join p1 p2 in
		match Pol.isIncl p pR with
		| Pol.Incl _ -> T.succeed state
		| Pol.NoIncl ->
			let err =
				Printf.sprintf "expected:\n%s\ngot:\n%s\n"
					(Pol.prExt varPr pR) (Pol.prExt varPr p)
			in
			T.fail name err state
	in
	T.suite "join" [
		T.suite "1" (List.map chk1 joinTcs);
		T.suite "2" (List.map chk2 joinTcs);
		T.suite "result" (List.map chkR joinTcs)
	]

let assignTcs: (string * Pol.assign_t * Pol.t * Pol.t) list
= [
	"no_x0", {Pol.var = x; Pol.lin = []; Pol.cst = Pol.S Nb.u},
		p [],
		p [eq [1, x] 1];

	"no_x1", {Pol.var = x; Pol.lin = [Nb.u, y]; Pol.cst = Pol.S Nb.z},
		p [],
		p [eq [1, x; -1, y] 0];

	"no_x2", {Pol.var = x; Pol.lin = [Nb.u, y]; Pol.cst = Pol.S Nb.z},
		p [le [1, y] 1],
		p [eq [1, x; -1, y] 0; le [1, y] 1];

	"no_x3", {Pol.var = z; Pol.lin = [Nb.u, x; Nb.u, y]; Pol.cst = Pol.S Nb.z},
		p [	le [-1, x; -1, y] (-7);
			le [-1, x; 1, y] 5],
		p [	le [-1, z] (-7);
			le [2, y; -1, z] 5;
			eq [1, x; 1, y; -1, z] 0];

	"x0_proj", {Pol.var = x; Pol.lin = []; Pol.cst = Pol.S Nb.u},
		p [le [1, x] 2],
		p [eq [1, x] 1];

	"x0", {Pol.var = x; Pol.lin = [Nb.u, x]; Pol.cst = Pol.S Nb.u},
		p [eq [1, x] 1],
		p [eq [-1, x] (-2)];

	"x1", {Pol.var = x; Pol.lin = [Nb.u, x]; Pol.cst = Pol.S Nb.u},
		p [le [1, x] 1],
		p [le [1, x] 2];

	"itv_proj_no_x0", {
		Pol.var = x;
		Pol.lin = [];
		Pol.cst = Pol.I {Pol.low = Pol.Infty; Pol.up = Pol.Infty}
	}, p [], p [];

	"itv_proj_x0", {
		Pol.var = x;
		Pol.lin = [];
		Pol.cst = Pol.I {Pol.low = Pol.Infty; Pol.up = Pol.Infty}
	}, p [le [1, x] 1], p [];

	"itv_proj_x1", {
		Pol.var = x;
		Pol.lin = [];
		Pol.cst = Pol.I {Pol.low = Pol.Infty; Pol.up = Pol.Infty}
	}, p [
		le [1, x] 1;
		le [-1, x; 1, y] 2
	], p [le [1, y] 3];

	"itv_up_ninv0", {
		Pol.var = x;
		Pol.lin = [];
		Pol.cst = Pol.I {Pol.low = Pol.Infty; Pol.up = Pol.Closed Nb.u}
	}, p [], p [le [1, x] 1];

	"itv_up_ninv1", {
		Pol.var = x;
		Pol.lin = [Nb.u, y];
		Pol.cst = Pol.I {Pol.low = Pol.Infty; Pol.up = Pol.Closed Nb.u}
	}, p [], p [le [1, x; -1, y] 1];

	"itv_up_ninv2", {
		Pol.var = x;
		Pol.lin = [];
		Pol.cst = Pol.I {Pol.low = Pol.Infty; Pol.up = Pol.Closed Nb.u}
	}, p [le [1, y] 1], p [
		le [1, x] 1;
		le [1, y] 1
	];

	"itv_low_ninv0", {
		Pol.var = x;
		Pol.lin = [];
		Pol.cst = Pol.I {Pol.low = Pol.Closed Nb.u; Pol.up = Pol.Infty}
	}, p [], p [le [-1, x] (-1)];

	"itv_up_ninv1", {
		Pol.var = x;
		Pol.lin = [Nb.u, y];
		Pol.cst = Pol.I {Pol.low = Pol.Closed Nb.u; Pol.up = Pol.Infty}
	}, p [], p [le [-1, x; 1, y] (-1)];

	"itv_up_ninv2", {
		Pol.var = x;
		Pol.lin = [];
		Pol.cst = Pol.I {Pol.low = Pol.Closed Nb.u; Pol.up = Pol.Infty}
	}, p [le [1, y] 1], p [
		le [-1, x] (-1);
		le [1, y] 1
	];

	"itv_bnd_ninv0", {
		Pol.var = x;
		Pol.lin = [];
		Pol.cst = Pol.I {Pol.low = Pol.Closed Nb.z; Pol.up = Pol.Closed Nb.u}
	}, p [], p [
		le [1, x] 1; 
		le [-1, x] 0
	];

	"itv_bnd_ninv1", {
		Pol.var = x;
		Pol.lin = [Nb.u, y];
		Pol.cst = Pol.I {Pol.low = Pol.Closed Nb.z; Pol.up = Pol.Closed Nb.u}
	}, p [], p [
		le [1, x; -1, y] 1; 
		le [-1, x; 1, y] 0
	];
]

(* Pol.assign *)
let assignTs: T.testT
=
	let chk (t, asg, p1, p) = fun state ->
		let (p2, _) = Pol.assign p1 asg in
		if Pol.isEqSyn p p2 then
			T.succeed state
		else
			let estr =
				"result not syntactically equal to expected" ^
				"\nresult:\n" ^ (Pol.prExt varPr p2) ^
				"\nexpected:\n" ^ (Pol.prExt varPr p)
			in
			T.fail t estr state
	in
	T.suite "assign" (List.map chk assignTcs)

(* Pol.assignM *)
let assignMTs: T.testT
=
	let chk (t, l, p1, p) = fun state ->
		let p2 = Pol.assignM p1 l in
		if Pol.isEqSyn p p2 then
			T.succeed state
		else
			let err =
				"result not syntactically equal to expected" ^
				"\nresult:\n" ^ (Pol.prExt varPr p2) ^
				"\nexpected:\n" ^ (Pol.prExt varPr p)
			in
			T.fail t err state
	in
	let tcs = [
		"noX0", [{ Pol.var = x; Pol.lin = []; Pol.cst = Pol.S Nb.u }], p [],
			p [ eq [1, x] 1 ];

		"noX1", [{ Pol.var = x; Pol.lin = [Nb.u, y]; Pol.cst = Pol.S Nb.z }],
			p [], p [ eq [1, x; -1, y] 0 ];

		"noX2", [{ Pol.var = x; Pol.lin = [Nb.u, y]; Pol.cst = Pol.S Nb.z }],
			p [
				le [1, y] 1
			], p [
				eq [1, x; -1, y] 0;
				le [1, y] 1 ];

		"xProj0", [{ Pol.var = x; Pol.lin = []; Pol.cst = Pol.S Nb.u }],
			p [
				le [1, x] 2
			], p [
				eq [1, x] 1 ] ;

		"x=F(x)0", [{ Pol.var = x; Pol.lin = [Nb.u, x]; Pol.cst = Pol.S Nb.u }],
			p [
				eq [1, x] 1
			], p [
				eq [-1, x] (-2) ];

		"x=F(x)1", [{ Pol.var = x; Pol.lin = [Nb.u, x]; Pol.cst = Pol.S Nb.u }],
			p [
				le [1, x] 1
			], p [
				le [1, x] 2 ];

		"swap0", [{ Pol.var = x; Pol.lin = [Nb.u, y]; Pol.cst = Pol.S Nb.z };
			{ Pol.var = y; Pol.lin = [Nb.u, x]; Pol.cst = Pol.S Nb.z }],
			p [], p [];

		"swap1", [{ Pol.var = x; Pol.lin = [Nb.u, y]; Pol.cst = Pol.S Nb.z };
			{ Pol.var = y; Pol.lin = [Nb.u, x]; Pol.cst = Pol.S Nb.z }],
			p [
				eq [1, x] 1
			], p [
				eq [1, y] 1 ];

		"swap2", [{ Pol.var = x; Pol.lin = [Nb.u, y]; Pol.cst = Pol.S Nb.z };
			{ Pol.var = y; Pol.lin = [Nb.u, x]; Pol.cst = Pol.S Nb.z }],
			p [
				eq [1, y] 1
			], p [
				eq [1, x] 1 ];
	] in
	T.suite "assignM" (List.map chk tcs)

(* Pol.assignRelSub *)
let assignRelSubTs: T.testT
=
	let chk (nm, p1, l, res)
	= fun st ->
		let maybeP2 = Pol.assignRelSub (Var.next e) p1 l in
		match res, maybeP2 with
		| None, None -> T.succeed st
		| None, Some _ -> T.fail nm "got Some _ instead of None" st
		| Some _, None -> T.fail nm "got None instead of Some _" st
		| Some p, Some p2 ->
			if Pol.isEqSyn p p2
			then T.succeed st
			else
				let err =
					"result not syntactically equal to expected" ^
					"\nresult:\n" ^ (Pol.prExt varPr p2) ^
					"\nexpected:\n" ^ (Pol.prExt varPr p)
				in
				T.fail nm err st
	in
	let o: Var.t -> Pol.relVarT
	= fun x -> Pol.OldSt x
	in
	let n: Var.t -> Pol.relVarT
	= fun x -> Pol.NewSt x
	in
	let mk': Cstr.cmpT -> (int * Pol.relVarT) list -> int -> Pol.relCstrT
	= fun cmp l cst -> {
		Pol.cmpSign = cmp;
		Pol.linear = List.map (fun (n, x) -> (Nb.mk1 n, x)) l;
		Pol.constant = Nb.mk1 cst}
	in
	let eq' = mk' Cstr.Eq in
	let le' = mk' Cstr.Le in
	let tcs: (string * Pol.t * Pol.relCstrT list * Pol.t option) list
	= [
	"triv0", p [], [], Some (p []);
	"triv1", p [eq [1, x] 0], [], Some (p [eq [1, x] 0]);
	"triv2", p [le [1, x] 0], [], Some (p [le [1, x] 0]);

	"id0", p [le [1, x] 2], [eq' [1, o x] 1], Some (p [le [1, x] 2]);
	"id1", p [le [1, x] 2], [eq' [0, n x] 0], Some (p [le [1, x] 2]);
	"id2", p [le [1, x] 2], [eq' [1, n x; -1, n x] 0], Some (p [le [1, x] 2]);

	"contr0", p [], [
			eq' [1, n x] 1;
			eq' [1, n x] 2
		], None;

	"non_inv0", p [], [eq' [1, n x] 0], Some (p [eq [1, x] 0]);
	"non_inv1", p [], [le' [1, n x] 0], Some (p [le [1, x] 0]);

	"inv0", p [], [eq' [1, n x; -1, o x] 0], Some (p []);
	"inv1", p [eq [1, x] 0], [eq' [1, n x; -1, o x] 1], Some (p [eq [1, x] 1]);
	"inv2", p [eq [1, x] 0], [le' [1, n x; -1, o x] 0], Some (p [le [1, x] 0]);
	] in
	T.suite "assignRelSub" (List.map chk tcs)

(* Pol.widen *)
let widenTs: T.testT
=
	let chk (t, p1, p2, pR) = fun state ->
		if Pol.isIncl p1 p2 = Pol.NoIncl then
			failwith "Pol_t.widen_ts"
		else
			let p = Pol.widen p1 p2 in
			if Pol.isEqSyn p pR then
				T.succeed state
			else
				let err =
					"p is different from expected pR" ^
					"\np:\n" ^ (Pol.prExt varPr p) ^ "\npR:\n" ^ (Pol.prExt varPr pR)
				in
				T.fail t err state
	in
	let tcs = [
		"halbwachs0", p [
			eq [1, y] 0;
			le [-1, x; 1, y] 0;
			le [1, x; -1, y] 1
		], p [
			le [-1, y] 0;
			le [-1, x; 1, y] 0;
			le [1, x; 1, y] 2
		], p [
			le [-1, y] 0;
			le [-1, x; 1, y] 0
		];

		"halbwachs1", p [
			eq [1, y] 0;
			le [-1, x] 0;
			le [1, x] 1
		], p [
			le [-1, y] 0;
			le [-1, x; 1, y] 0;
			le [1, x; 1, y] 2
		], p [
			le [-1, y] 0;
			le [-1, x; 1, y] 0
		];


		(*"dots0", p [
			eq [1, x] 0;
			eq [1, y] 0
		], p [
			eq [2, x; -1, y] 0;
			le [1, y] 2;
			le [-1, y] 0
		], p [
			eq [2, x; -1, y] 0;
			le [-1, y] 0
		]*)
	] in
	T.suite "widen" (List.map chk tcs)

(* Pol.itvize *)
let itvizeTcs: (string * Vec.t * Pol.itvT * Pol.t) list
= [
	"nil0", Vec.mk [Nb.u, x], {Pol.low = Pol.Infty; Pol.up = Pol.Infty}, p [];
	"direct0", Vec.mk [Nb.u, x], {Pol.low = Pol.Infty; Pol.up = Pol.Closed (Nb.mk1 2)}, p [
		le [1, x] 2 ];

	"direct1", Vec.mk [Nb.u, x], {Pol.low = Pol.Infty; Pol.up = Pol.Open (Nb.mk1 2)}, p [
		lt [1, x] 2 ];

	"direct2", Vec.mk [Nb.u, x], {Pol.low = Pol.Closed (Nb.mk1 (-2)); Pol.up = Pol.Infty}, p [
		le [-1, x] 2 ];

	"direct3", Vec.mk [Nb.u, x], {Pol.low = Pol.Open (Nb.mk1 (-2)); Pol.up = Pol.Infty}, p [
		lt [-1, x] 2 ];

	"direct4", Vec.mk [Nb.u, x], {Pol.low = Pol.Closed (Nb.mk1 2); Pol.up = Pol.Closed (Nb.mk1 2)}, p [
		eq [1, x] 2 ];

	"comb0", Vec.mk [Nb.u, x], {Pol.low = Pol.Infty; Pol.up = Pol.Closed (Nb.mk1 2)}, p [
		le [1, y] 2;
		le [2, x; -1, y] 2 ];

	"comb1", Vec.mk [Nb.u, x], {Pol.low = Pol.Infty; Pol.up = Pol.Open (Nb.mk1 2)}, p [
		lt [1, y] 2;
		le [2, x; -1, y] 2 ];

	"comb2", Vec.mk [Nb.u, x], {Pol.low = Pol.Infty; Pol.up = Pol.Open (Nb.mk1 2)}, p [
		le [1, y] 2;
		lt [2, x; -1, y] 2 ];

	"eq0", Vec.mk [Nb.u, x], {Pol.low = Pol.Infty; Pol.up = Pol.Closed Nb.z}, p [
		eq [1, y] 2;
		le [1, x; 1, y] 2 ]
]

(* XXX: certificates should be syntactically checked *)
let itvizeTs: T.testT
=
	let chk (name, v, itv, p) = fun state ->
		let (itv1, _, _) = Pol.itvize p v in
		if eqItv itv itv1 then
			T.succeed state
		else
			T.fail name (Pol.prItv varPr v itv1) state
	in
	T.suite "itvize" (List.map chk itvizeTcs)

(* Pol.rename *)
let renameTs : T.testT
=	let chk : string * Var.t * Var.t * Pol.t * Pol.t option -> T.stateT -> T.stateT
	= fun (nm, x, y, p, op) st ->
		match op with
		| Some p' ->
			if Pol.isEqSyn (Pol.rename x y p) p'
			then T.succeed st
			else
				let e : string = "failed" in
				T.fail nm e st
		| None -> try
			let _ = Pol.rename x y p in
			let e : string = "failed" in
			T.fail nm e st
			with Assert_failure (_, _, _) -> T.succeed st
	in
	let tcs : (string * Var.t * Var.t * Pol.t * Pol.t option) list
	= [
		"nil0", x, y, p [], Some (p []);

		"ok0", x, y, p [le [1, x] 0], Some (p [le [1, y] 0]);

		"ok1", x, y, p [eq [1, x] 0], Some (p [eq [1, y] 0]);

		"ko0", x, y, p [le [1, y] 0], None;

		"ko1", x, y, p [eq [1, y] 0], None
	] in
	T.suite "rename" (List.map chk tcs)

let ts: T.testT
= T.suite "Pol" [
	certCombineFwdTs;
	addTs;
	meetTs;
	isInclTs;
	projectTs;
	projectMTs;
	joinSetupTs;
	joinTs;
	assignTs;
	assignMTs;
	assignRelSubTs;
	widenTs;
	itvizeTs;
	renameTs
]
