(* Rtree.set *)
let setNilConsTs: T.testT
=
	let chk (name, tree) = fun state ->
		match tree with
		| Rtree.Nil -> T.fail name "rt is nil" state
		| Rtree.Sub _ -> T.succeed state
	in
	let tcs = [
		"_", Rtree.set 0 Rtree.Nil Var.XH 1;
		"O_", Rtree.set 0 Rtree.Nil (Var.XO Var.XH) 1;
		"I_", Rtree.set 0 Rtree.Nil (Var.XI Var.XH) 1;
		"OO_", Rtree.set 0 Rtree.Nil (Var.XO (Var.XO Var.XH)) 1;
		"OI_", Rtree.set 0 Rtree.Nil (Var.XO (Var.XI Var.XH)) 1;
		"IO_", Rtree.set 0 Rtree.Nil (Var.XI (Var.XO Var.XH)) 1;
		"II_", Rtree.set 0 Rtree.Nil (Var.XI (Var.XI Var.XH)) 1;
	] in
	T.suite "nilCons" (List.map chk tcs)

let setOrderTs: T.testT
=
	let chk (name, tree1, tree2) = fun state ->
		if tree1 = tree2 then
			T.succeed state
		else
			T.fail name "not equal" state
	in
	let set = Rtree.set 0 in
	let tcs = [
		"_/O_",
			set (set Rtree.Nil Var.XH 1) (Var.XO Var.XH) 2,
			set (set Rtree.Nil (Var.XO Var.XH) 2) Var.XH 1;
		"_/I_",
			set (set Rtree.Nil Var.XH 1) (Var.XI Var.XH) 2,
			set (set Rtree.Nil (Var.XI Var.XH) 2) Var.XH 1;
		"O_/I_",
			set (set Rtree.Nil (Var.XO Var.XH) 1) (Var.XI Var.XH) 2,
			set (set Rtree.Nil (Var.XI Var.XH) 2) (Var.XO Var.XH) 1;
		"_/OO_",
			set (set Rtree.Nil Var.XH 1) (Var.XO (Var.XO Var.XH)) 2,
			set (set Rtree.Nil (Var.XO (Var.XO Var.XH)) 2) Var.XH 1;
		"_/OI_",
			set (set Rtree.Nil Var.XH 1) (Var.XO (Var.XI Var.XH)) 2,
			set (set Rtree.Nil (Var.XO (Var.XI Var.XH)) 2) Var.XH 1;
		"_/IO_",
			set (set Rtree.Nil Var.XH 1) (Var.XI (Var.XO Var.XH)) 2,
			set (set Rtree.Nil (Var.XI (Var.XO Var.XH)) 2) Var.XH 1;
		"_/II_",
			set (set Rtree.Nil Var.XH 1) (Var.XI (Var.XI Var.XH)) 2,
			set (set Rtree.Nil (Var.XI (Var.XI Var.XH)) 2) Var.XH 1
	] in
	T.suite "order" (List.map chk tcs)

let setTs: T.testT
= T.suite "set" [setNilConsTs; setOrderTs]

(* Rtree.find *)
let findTs: T.testT
=
	let chk (name, predicate, tree, expected) = fun state ->
		let pr
		= function
			| None -> "None"
			| Some _ -> "Some _"
		in
		let actual = Rtree.find predicate tree in
		if List.exists ((=) actual) expected then
			T.succeed state
		else
			T.fail name (pr actual) state
	in
	let tcs = [
		"nil0", (fun _ -> Some 0), Rtree.Nil, [None];
		"nil1", (fun _ -> None), Rtree.Nil, [None];
		"yes0", (fun _ -> Some 0), Rtree.mk 0 [Var.XH, 1], [Some (Var.XH, 0)];
		"yes1", (fun _ -> Some 0), Rtree.mk 0 [Var.XH, 1; Var.XO Var.XH, 2],
			[Some (Var.XH, 0); Some (Var.XO Var.XH, 0)];

		"no0", (fun _ -> None), Rtree.mk 0 [Var.XH, 1], [None];
		"no1", (fun i -> if i = 0 then Some 0 else None), Rtree.mk 0 [Var.XH, 1; Var.XO Var.XH, 2], [None];
		"ret0", (fun i -> Some i), Rtree.mk 0 [Var.XH, 1; Var.XO Var.XH, 2],
			[Some (Var.XH, 1); Some (Var.XO Var.XH, 2)];
		"ret0", (fun i -> if i = 1 then Some 10 else None), Rtree.mk 0 [Var.XH, 1; Var.XI Var.XH, 2],
			[Some (Var.XH, 10)]
	] in
	T.suite "find" (List.map chk tcs)

(* Rtree.pathsGet *)
let pathsGetTs: T.testT
=
	let chk (name, tree, paths) = fun state ->
		let res = Rtree.pathsGet tree in
		if res = paths then
			T.succeed state
		else
			T.fail name "not equal" state
	in
	let tcs = [
		"nil", Rtree.Nil, [];
		"unit0", Rtree.Sub (Rtree.Nil, true, Rtree.Nil), [Var.XH];
		"unit1", Rtree.Sub (Rtree.Nil, false, Rtree.Nil), [];
		"unit2", Rtree.Sub (Rtree.Sub (Rtree.Nil, true, Rtree.Nil), false, Rtree.Nil), [Var.XO Var.XH];
		"unit3", Rtree.Sub (Rtree.Nil, false, Rtree.Sub (Rtree.Nil, true, Rtree.Nil)), [Var.XI Var.XH];
		"m0", Rtree.Sub (Rtree.Sub (Rtree.Nil, true, Rtree.Nil), true, Rtree.Nil), [Var.XH; Var.XO Var.XH];
		"m1", Rtree.Sub (Rtree.Nil, true, Rtree.Sub (Rtree.Nil, true, Rtree.Nil)), [Var.XH; Var.XI Var.XH]
	] in
	T.suite "pathsGet" (List.map chk tcs)

(* Rtree.toList *)
let toListTs: T.testT
=
	let chk (name, tree, paths) = fun state ->
		let res = Rtree.toList tree in
		if res = paths then
			T.succeed state
		else
			T.fail name "not equal" state
	in
	let tcs = [
		"nil", Rtree.Nil, [];
		"unit0", Rtree.Sub (Rtree.Nil, true, Rtree.Nil), [true, Var.XH];
		"unit1", Rtree.Sub (Rtree.Nil, false, Rtree.Nil), [false, Var.XH];
		"unit2", Rtree.Sub (Rtree.Sub (Rtree.Nil, true, Rtree.Nil), false, Rtree.Nil), [false, Var.XH; true, Var.XO Var.XH];
		"unit3", Rtree.Sub (Rtree.Nil, false, Rtree.Sub (Rtree.Nil, true, Rtree.Nil)), [false, Var.XH; true, Var.XI Var.XH];
		"m0", Rtree.Sub (Rtree.Sub (Rtree.Nil, true, Rtree.Nil), true, Rtree.Nil), [true, Var.XH; true, Var.XO Var.XH];
		"m1", Rtree.Sub (Rtree.Nil, true, Rtree.Sub (Rtree.Nil, true, Rtree.Nil)), [true, Var.XH; true, Var.XI Var.XH]
	] in
	T.suite "toList" (List.map chk tcs)

let ts: T.testT
= T.suite "Rtree" [setTs; findTs; pathsGetTs; toListTs]

