(* Var.next *)
let nextTs: T.testT
=
	let chk (name, var, expected) = fun state ->
		if Var.next var = expected then
			T.succeed state
		else
			T.fail name "not equal" state
	in
	let tcs: (string * Var.t * Var.t) list = [
	"nil", Var.XH, Var.XO Var.XH;
	"cons0", Var.XO Var.XH, Var.XI Var.XH;
	"cons1", Var.XI Var.XH, Var.XO (Var.XO Var.XH);
	"rec0", Var.XO (Var.XO (Var.XI (Var.XI Var.XH))),
		Var.XO (Var.XI (Var.XO (Var.XO Var.XH)));
	"rec1", Var.XI (Var.XI (Var.XI (Var.XI Var.XH))),
		Var.XO (Var.XO (Var.XO (Var.XO (Var.XO Var.XH))));
	] in
	T.suite "next" (List.map chk tcs)

(* Var.cmp *)
let cmpTs: T.testT
=
	let chk (name, ans, x1, x2) = fun state ->
		let res = Var.cmp x1 x2 in
		if ans = res then
			T.succeed state
		else
			T.fail name (string_of_int res) state
	in
	let tcs = [
		"eq0", 0, Var.XH, Var.XH;
		"eq1", 0, Var.XO Var.XH, Var.XO Var.XH;
		"eq2", 0, Var.XI Var.XH, Var.XI Var.XH;
		"eq3", 0, Var.XO (Var.XI Var.XH), Var.XO (Var.XI Var.XH);
		"eq4", 0, Var.XI (Var.XI Var.XH), Var.XI (Var.XI Var.XH);
		"eq5", 0, Var.XO (Var.XO Var.XH), Var.XO (Var.XO Var.XH);
		"eq6", 0, Var.XI (Var.XO Var.XH), Var.XI (Var.XO Var.XH);

		"lt0", -1, Var.XH, Var.XO Var.XH;
		"lt1", -1, Var.XH, Var.XI Var.XH;
		"lt2", -1, Var.XO Var.XH, Var.XI Var.XH;
		"lt3", -1, Var.XO Var.XH, Var.XO (Var.XO Var.XH);
		"lt4", -1, Var.XO Var.XH, Var.XI (Var.XO Var.XH);

		"gt0", 1, Var.XO Var.XH, Var.XH;
		"gt1", 1, Var.XI Var.XH, Var.XH;
		"gt2", 1, Var.XO (Var.XO Var.XH), Var.XH;
		"gt3", 1, Var.XO (Var.XI Var.XH), Var.XO (Var.XO Var.XH);
		"gt4", 1, Var.XO (Var.XO (Var.XI Var.XH)), Var.XO (Var.XO Var.XH)
	] in
	T.suite "cmp" (List.map chk tcs)

(* Var.horizon *)
let horizonTs
=
	let [a; b; c; d; e; f; g; h; i; j; k; l] = T.pathGen 12 in
	let chk (name, h, vars) state =
		let h1 = Var.horizon vars in
		if h = h1 then
			T.succeed state
		else
			T.fail name "not equal" state
	in
	let tcs = [
		"empty", a, [];
		"one0", b, [a];
		"one1", g, [f];
		"ordered0", f, [c; d; e];
		"ordered1", l, [a; d; j; k];
		"gen0", j, [b; a; g; i; c]
	] in
	T.suite "horizon" (List.map chk tcs)

let intConvTcs = [
	"one", 1, Var.XH;
	"two", 2, Var.XO Var.XH;
	"three", 3, Var.XI Var.XH;
	"four", 4, Var.XO (Var.XO Var.XH)
]

(* Var.toInt *)
let toIntTs: T.testT
= (* XXX: test the overflow case *)
	let chk (name, i, p) = fun state ->
		let res = Var.toInt p in
		if i = res then
			T.succeed state
		else
			T.fail name "not equal" state
	in
	T.suite "toInt" (List.map chk intConvTcs)

(* Var.fromInt *)
let fromIntTs: T.testT
=
	let chk (name, i, p) state =
		let res = Var.fromInt i in
		if res = p then
			T.succeed state
		else
			T.fail name "not equal" state
	in
	T.suite "fromInt" (List.map chk intConvTcs)

let ts: T.testT
= T.suite "Var" [nextTs; cmpTs; horizonTs; toIntTs; fromIntTs]

