type sign_t = P | Z | N

let sign a = if a < 0 then N else if a > 0 then P else Z
let pr = function P -> "P" | Z -> "Z" | N -> "N"

(* Nb.Z.ofString *)
let ofStringTs : T.testT
=	let chk : (string * string * Nb.Z.t) -> T.stateT -> T.stateT
	= fun (nm, s, r) st ->
		let n = Nb.Z.ofString s in
		if Nb.Z.cmp r n = 0
		then T.succeed st
		else
			let e = Nb.Z.pr n in
			T.fail nm e st
	in
	let tcs : (string * string * Nb.Z.t) list
	= [
		"z", "0", Nb.Z.z;
		"pz", "+0", Nb.Z.z;
		"nz", "-0", Nb.Z.z;
		"u", "1", Nb.Z.u;
		"pu", "+1", Nb.Z.u;
		"nu", "-1", Nb.Z.mk ~- 1;
		"10", "10", Nb.Z.mk 10
	] in
	List.map chk tcs |> T.suite "ofString"

(* Nb.Z.add *)
let addTs : T.testT
=	let chk : (string * Nb.Z.t * Nb.Z.t * Nb.Z.t) -> T.stateT -> T.stateT
	= fun (nm, a, b, r) st ->
		let c = Nb.Z.add a b in
		if Nb.Z.cmp r c = 0
		then T.succeed st
		else
			let e = Nb.Z.pr c in
			T.fail nm e st
	in
	let tcs : (string * Nb.Z.t * Nb.Z.t * Nb.Z.t) list
	= [
		"zz", Nb.Z.z, Nb.Z.z, Nb.Z.z;
		"zu", Nb.Z.z, Nb.Z.u, Nb.Z.u;
		"uu", Nb.Z.u, Nb.Z.u, Nb.Z.mk 2;
		"np", Nb.Z.mk ~- 2, Nb.Z.u, Nb.Z.mk ~- 1;
		"2,2", Nb.Z.mk 2, Nb.Z.mk 2, Nb.Z.mk 4
	] in
	List.map chk tcs |> T.suite "add"

(* Nb.Z.sub *)
let subTs : T.testT
=	let chk : (string * Nb.Z.t * Nb.Z.t * Nb.Z.t) -> T.stateT -> T.stateT
	= fun (nm, a, b, r) st ->
		let c = Nb.Z.sub a b in
		if Nb.Z.cmp r c = 0
		then T.succeed st
		else
			let e = Nb.Z.pr c in
			T.fail nm e st
	in
	let tcs : (string * Nb.Z.t * Nb.Z.t * Nb.Z.t) list
	= [
		"zz", Nb.Z.z, Nb.Z.z, Nb.Z.z;
		"zu", Nb.Z.z, Nb.Z.u, Nb.Z.mk ~- 1;
		"uu", Nb.Z.u, Nb.Z.u, Nb.Z.z;
		"np", Nb.Z.mk ~- 2, Nb.Z.u, Nb.Z.mk ~- 3;
		"2,1", Nb.Z.mk 2, Nb.Z.mk 1, Nb.Z.mk 1
	] in
	List.map chk tcs |> T.suite "sub"

(* Nb.Z.mul *)
let mulTs : T.testT
=	let chk : (string * Nb.Z.t * Nb.Z.t * Nb.Z.t) -> T.stateT -> T.stateT
	= fun (nm, a, b, r) st ->
		let c = Nb.Z.mul a b in
		if Nb.Z.cmp r c = 0
		then T.succeed st
		else
			let e = Nb.Z.pr c in
			T.fail nm e st
	in
	let tcs : (string * Nb.Z.t * Nb.Z.t * Nb.Z.t) list
	= [
		"zz", Nb.Z.z, Nb.Z.z, Nb.Z.z;
		"zu", Nb.Z.z, Nb.Z.u, Nb.Z.z;
		"uu", Nb.Z.u, Nb.Z.u, Nb.Z.u;
		"np", Nb.Z.mk ~- 2, Nb.Z.u, Nb.Z.mk ~- 2;
		"2,1", Nb.Z.mk 2, Nb.Z.mk 1, Nb.Z.mk 2
	] in
	List.map chk tcs |> T.suite "mul"

(* Nb.Z.div *)
let divTs : T.testT
=	let chk : (string * Nb.Z.t * Nb.Z.t * Nb.Z.t) -> T.stateT -> T.stateT
	= fun (nm, a, b, r) st ->
		let c = Nb.Z.div a b in
		if Nb.Z.cmp r c = 0
		then T.succeed st
		else
			let e = Nb.Z.pr c in
			T.fail nm e st
	in
	let tcs : (string * Nb.Z.t * Nb.Z.t * Nb.Z.t) list
	= [
		"zu", Nb.Z.z, Nb.Z.u, Nb.Z.z;
		"uu", Nb.Z.u, Nb.Z.u, Nb.Z.u;
		"np", Nb.Z.mk ~- 2, Nb.Z.u, Nb.Z.mk ~- 2;
		"2,2", Nb.Z.mk 2, Nb.Z.mk 2, Nb.Z.mk 1;
		"3,2", Nb.Z.mk 3, Nb.Z.mk 2, Nb.Z.mk 1
	] in
	List.map chk tcs |> T.suite "div"

(* Nb.Z.rem *)
let remTs : T.testT
=	let chk : (string * Nb.Z.t * Nb.Z.t * Nb.Z.t) -> T.stateT -> T.stateT
	= fun (nm, a, b, r) st ->
		let c = Nb.Z.rem a b in
		if Nb.Z.cmp r c = 0
		then T.succeed st
		else
			let e = Nb.Z.pr c in
			T.fail nm e st
	in
	let tcs : (string * Nb.Z.t * Nb.Z.t * Nb.Z.t) list
	= [
		"zu", Nb.Z.z, Nb.Z.u, Nb.Z.z;
		"uu", Nb.Z.u, Nb.Z.u, Nb.Z.z;
		"np", Nb.Z.mk ~- 3, Nb.Z.mk 2, Nb.Z.mk ~- 1;
		"3,2", Nb.Z.mk 3, Nb.Z.mk 2, Nb.Z.mk 1
	] in
	List.map chk tcs |> T.suite "rem"

(* Nb.Z.cmp *)
let cmpTs: T.testT
=
	let chk (name, n1, n2, expected) = fun state ->
		let s = sign (Nb.Z.cmp n1 n2) in
		if s = expected then
			T.succeed state
		else
			T.fail name (pr s) state
	in
	let tcs = [
		"zz0", Nb.Z.z, Nb.Z.z, Z;
		"zu0", Nb.Z.z, Nb.Z.u, N;
		"uz0", Nb.Z.u, Nb.Z.z, P;
		"pz0", Nb.Z.mk 2, Nb.Z.z, P;
		"zp0", Nb.Z.z, Nb.Z.mk 2, N;
		"nz0", Nb.Z.mk (-2), Nb.Z.z, N;
		"zn0", Nb.Z.z, Nb.Z.mk (-2), P
	] in
	T.suite "cmp" (List.map chk tcs)

(* Nb.Z.neg *)
let negTs: T.testT
=
	let chk (name, n, expected) = fun state ->
		let negN = Nb.Z.neg n in
		if Nb.Z.cmp negN expected = 0 then
			T.succeed state
		else
			T.fail name (Nb.Z.pr negN) state
	in
	let tcs = [
		"z0", Nb.Z.z, Nb.Z.z;
		"u0", Nb.Z.u, Nb.Z.mk (-1);
		"u1", Nb.Z.mk (-1), Nb.Z.u;
		"p0", Nb.Z.mk 2, Nb.Z.mk (-2);
		"n0", Nb.Z.mk (-2), Nb.Z.mk 2
	] in
	T.suite "neg" (List.map chk tcs)

(* Nb.Z.lAnd *)
let lAndTs: T.testT
=
	let chkRes (name, n1, n2, expected) = fun state ->
		let r = Nb.Z.lAnd n1 n2 in
		if Nb.Z.cmp r expected = 0 then
			T.succeed state
		else
			T.fail name (Nb.Z.pr r) state
	in
	let chkSym (name, n1, n2, _) = fun state ->
		let r1 = Nb.Z.lAnd n1 n2 in
		let r2 = Nb.Z.lAnd n2 n1 in
		if Nb.Z.cmp r1 r2 = 0 then
			T.succeed state
		else
			let err = (Nb.Z.pr r1) ^ " vs. " ^ (Nb.Z.pr r2) in
			T.fail name err state
	in
	let tcs = [
		"z0", Nb.Z.z, Nb.Z.z, Nb.Z.z;
		"z1", Nb.Z.u, Nb.Z.z, Nb.Z.z;
		"z2", Nb.Z.mk (-2), Nb.Z.z, Nb.Z.z;
		"u0", Nb.Z.mk 2, Nb.Z.u, Nb.Z.z;
		"u1", Nb.Z.mk 3, Nb.Z.u, Nb.Z.u;
		"u2", Nb.Z.u, Nb.Z.u, Nb.Z.u;
	] in
	T.suite "lAnd" [
		T.suite "res" (List.map chkRes tcs);
		T.suite "sym" (List.map chkSym tcs)
	]

(* Nb.Z.shiftR *)
let shiftRTs: T.testT =
	let chk (name, n, nb, expected) = fun state ->
		let r = Nb.Z.shiftR n nb in
		if Nb.Z.cmp r expected = 0 then
			T.succeed state
		else
			T.fail name (Nb.Z.pr r) state
	in
	let tcs = [
		"z0", Nb.Z.z, 0, Nb.Z.z;
		"z1", Nb.Z.z, 1, Nb.Z.z;
		"u0", Nb.Z.u, 0, Nb.Z.u;
		"u1", Nb.Z.u, 1, Nb.Z.z;
		"u2", Nb.Z.u, 2, Nb.Z.z;
		"p0", Nb.Z.mk 2, 0, Nb.Z.mk 2;
		"p1", Nb.Z.mk 2, 1, Nb.Z.u;
		"p2", Nb.Z.mk 2, 2, Nb.Z.z;
		"p3", Nb.Z.mk 3, 1, Nb.Z.u;
		"n0", Nb.Z.mk (-1), 1, Nb.Z.mk (-1);
		"n1", Nb.Z.mk (-2), 1, Nb.Z.mk (-1);
		"n2", Nb.Z.mk (-3), 1, Nb.Z.mk (-2);
	] in
	T.suite "shiftR" (List.map chk tcs)

let ts: T.testT
= T.suite "Nb.Z" [
	ofStringTs;
	addTs;
	subTs;
	mulTs;
	divTs;
	remTs;
	cmpTs; negTs; lAndTs; shiftRTs]
